# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'


import io
import time
import os
import re
import threading
import math
try:
	import queue
except ImportError:
	import Queue as queue

# noinspection PyCompatibility
from past.builtins import basestring

from serial import SerialTimeoutException

from octoprint.settings import settings
from octoprint.plugin import plugin_manager
from octoprint.util import RepeatedTimer, monotonic_time, to_bytes, to_unicode

from typing import Any


# noinspection PyBroadException
class VirtualPrinter(object):
	command_regex = re.compile(r"^([GMTF])(\d+)")
	sleep_regex = re.compile(r"sleep (\d+)")
	sleep_after_regex = re.compile(r"sleep_after ([GMTF]\d+) (\d+)")
	sleep_after_next_regex = re.compile(r"sleep_after_next ([GMTF]\d+) (\d+)")
	custom_action_regex = re.compile(r"action_custom ([a-zA-Z0-9_]+)(\s+.*)?")
	prepare_ok_regex = re.compile(r"prepare_ok (.*)")
	send_regex = re.compile(r"send (.*)")
	set_ambient_regex = re.compile(r"set_ambient ([-+]?[0-9]*\.?[0-9]+)")
	start_sd_regex = re.compile(r"start_sd (.*)")
	select_sd_regex = re.compile(r"select_sd (.*)")

	def __init__(self, seriallog_handler=None, read_timeout=5.0, write_timeout=10.0):
		import logging
		self._logger = logging.getLogger("octoprint.plugins.virtual_printer.VirtualPrinter")

		self._seriallog = logging.getLogger("octoprint.plugin.virtual_printer.VirtualPrinter.serial")
		self._seriallog.setLevel(logging.CRITICAL)
		self._seriallog.propagate = False

		if seriallog_handler is not None:
			import logging.handlers
			self._seriallog.addHandler(seriallog_handler)
			self._seriallog.setLevel(logging.INFO)

		self._seriallog.info(u"-"*78)

		self._read_timeout = read_timeout
		self._write_timeout = write_timeout

		self._rx_buffer_size = settings().getInt(["devel", "virtualPrinter", "rxBuffer"])

		self.incoming = CharCountingQueue(self._rx_buffer_size, name="RxBuffer")
		self.outgoing = queue.Queue()
		self.buffered = queue.Queue(maxsize=settings().getInt(["devel", "virtualPrinter", "commandBuffer"]))

		if settings().getBoolean(["devel", "virtualPrinter", "simulateReset"]):
			for item in settings().get(["devel", "virtualPrinter", "resetLines"]):
				self._send(item + "\n")

		self._prepared_oks = []
		prepared = settings().get(["devel", "virtualPrinter", "preparedOks"])
		if prepared and isinstance(prepared, list):
			for prep in prepared:
				self._prepared_oks.append(prep)

		self._prepared_errors = []

		self._errors = settings().get(["devel", "virtualPrinter", "errors"], merged=True)

		self.currentExtruder = 0
		self.extruderCount = settings().getInt(["devel", "virtualPrinter", "numExtruders"])
		self.pinnedExtruders = settings().get(["devel", "virtualPrinter", "pinnedExtruders"])
		if self.pinnedExtruders is None:
			self.pinnedExtruders = dict()
		self.sharedNozzle = settings().getBoolean(["devel", "virtualPrinter", "sharedNozzle"])
		self.temperatureCount = (1 if self.sharedNozzle else self.extruderCount)

		self._ambient_temperature = settings().getFloat(["devel", "virtualPrinter", "ambientTemperature"])

		self.temp = [self._ambient_temperature] * self.temperatureCount
		self.targetTemp = [0.0] * self.temperatureCount
		self.bedTemp = self._ambient_temperature
		self.bedTargetTemp = 0.0
		self.chamberTemp = self._ambient_temperature
		self.chamberTargetTemp = 0.0
		self.lastTempAt = monotonic_time()

		self._relative = True
		self._lastX = 0.0
		self._lastY = 0.0
		self._lastZ = 0.0
		self._lastE = [0.0] * self.extruderCount
		self._lastF = 200

		self._unitModifier = 1
		self._feedrate_multiplier = 100
		self._flowrate_multiplier = 100

		self._virtualSd = settings().getBaseFolder("virtualSd")
		self._sdCardReady = True
		self._sdPrinter = None
		self._sdPrintingSemaphore = threading.Event()
		self._selectedSdFile = None
		self._selectedSdFileSize = None
		self._selectedSdFilePos = None

		self._writingToSd = False
		self._writingToSdHandle = None
		self._newSdFilePos = None

		self._heatingUp = False

		self._okBeforeCommandOutput = settings().getBoolean(["devel", "virtualPrinter", "okBeforeCommandOutput"])
		self._supportM112 = settings().getBoolean(["devel", "virtualPrinter", "supportM112"])
		self._supportF = settings().getBoolean(["devel", "virtualPrinter", "supportF"])

		self._sendWait = settings().getBoolean(["devel", "virtualPrinter", "sendWait"])
		self._sendBusy = settings().getBoolean(["devel", "virtualPrinter", "sendBusy"])
		self._waitInterval = settings().getFloat(["devel", "virtualPrinter", "waitInterval"])
		self._busyInterval = settings().getFloat(["devel", "virtualPrinter", "busyInterval"])

		self._echoOnM117 = settings().getBoolean(["devel", "virtualPrinter", "echoOnM117"])

		self._brokenM29 = settings().getBoolean(["devel", "virtualPrinter", "brokenM29"])
		self._brokenResend = settings().getBoolean(["devel", "virtualPrinter", "brokenResend"])

		self._m115FormatString = settings().get(["devel", "virtualPrinter", "m115FormatString"])
		self._firmwareName = settings().get(["devel", "virtualPrinter", "firmwareName"])

		self._okFormatString = settings().get(["devel", "virtualPrinter", "okFormatString"])

		self._capabilities = settings().get(["devel", "virtualPrinter", "capabilities"], merged=True)

		self._temperature_reporter = None
		self._sdstatus_reporter = None

		self.current_line = 0
		self.lastN = 0

		self._incoming_lock = threading.RLock()

		self._debug_awol = False
		self._debug_sleep = 0
		self._sleepAfterNext = dict()
		self._sleepAfter = dict()
		self._rerequest_last = False

		self._dont_answer = False

		self._debug_drop_connection = False

		self._action_hooks = plugin_manager().get_hooks("octoprint.plugin.virtual_printer.custom_action")

		self._killed = False

		self._triggerResendAt100 = True
		self._triggerResendWithTimeoutAt105 = True
		self._triggerResendWithMissingLinenoAt110 = True
		self._triggerResendWithChecksumMismatchAt115 = True

		readThread = threading.Thread(target=self._processIncoming, name="octoprint.plugins.virtual_printer.wait_thread")
		readThread.start()

		bufferThread = threading.Thread(target=self._processBuffer, name="octoprint.plugins.virtual_printer.buffer_thread")
		bufferThread.start()

	def __str__(self):
		return "VIRTUAL(read_timeout={read_timeout},write_timeout={write_timeout},options={options})"\
			.format(read_timeout=self._read_timeout, write_timeout=self._write_timeout, options=settings().get(["devel", "virtualPrinter"]))

	def _reset(self):
		with self._incoming_lock:
			self._relative = True
			self._lastX = 0.0
			self._lastY = 0.0
			self._lastZ = 0.0
			self._lastE = [0.0] * self.extruderCount
			self._lastF = 200

			self._unitModifier = 1
			self._feedrate_multiplier = 100
			self._flowrate_multiplier = 100

			self._sdCardReady = True
			self._sdPrinting = False
			if self._sdPrinter:
				self._sdPrinting = False
				self._sdPrintingSemaphore.set()
			self._sdPrinter = None
			self._selectedSdFile = None
			self._selectedSdFileSize = None
			self._selectedSdFilePos = None

			if self._writingToSdHandle:
				try:
					self._writingToSdHandle.close()
				except Exception:
					pass
			self._writingToSd = False
			self._writingToSdHandle = None
			self._newSdFilePos = None

			self._heatingUp = False

			self.current_line = 0
			self.lastN = 0

			self._debug_awol = False
			self._debug_sleep = 0
			self._sleepAfterNext.clear()
			self._sleepAfter.clear()

			self._dont_answer = False

			self._debug_drop_connection = False

			self._killed = False

			self._triggerResendAt100 = True
			self._triggerResendWithTimeoutAt105 = True
			self._triggerResendWithMissingLinenoAt110 = True
			self._triggerResendWithChecksumMismatchAt115 = True

			if self._temperature_reporter is not None:
				self._temperature_reporter.cancel()
				self._temperature_reporter = None

			if self._sdstatus_reporter is not None:
				self._sdstatus_reporter.cancel()
				self._sdstatus_reporter = None

			self._clearQueue(self.incoming)
			self._clearQueue(self.outgoing)
			self._clearQueue(self.buffered)

			if settings().getBoolean(["devel", "virtualPrinter", "simulateReset"]):
				for item in settings().get(["devel", "virtualPrinter", "resetLines"]):
					self._send(item + "\n")

	@property
	def timeout(self):
		return self._read_timeout

	@timeout.setter
	def timeout(self, value):
		self._logger.debug("Setting read timeout to {}s".format(value))
		self._read_timeout = value

	@property
	def write_timeout(self):
		return self._write_timeout

	@write_timeout.setter
	def write_timeout(self, value):
		self._logger.debug("Setting write timeout to {}s".format(value))
		self._write_timeout = value

	# noinspection PyMethodMayBeStatic
	def _clearQueue(self, q):
		try:
			while q.get(block=False):
				q.task_done()
				continue
		except queue.Empty:
			pass

	def _processIncoming(self):
		next_wait_timeout = monotonic_time() + self._waitInterval
		buf = b""
		while self.incoming is not None and not self._killed:
			self._simulateTemps()

			if self._heatingUp:
				time.sleep(1)
				continue

			try:
				data = self.incoming.get(timeout=0.01)
				data = to_bytes(data, encoding="ascii", errors="replace")
				self.incoming.task_done()
			except queue.Empty:
				if self._sendWait and monotonic_time() > next_wait_timeout:
					self._send("wait")
					next_wait_timeout = monotonic_time() + self._waitInterval
				continue
			except Exception:
				if self.incoming is None:
					# just got closed
					break

			if data is not None:
				buf += data
				nl = buf.find(b"\n")+1
				if nl > 0:
					data = buf[:nl]
					buf = buf[nl:]
				else:
					continue

			next_wait_timeout = monotonic_time() + self._waitInterval

			if data is None:
				continue

			if self._dont_answer:
				self._dont_answer = False
				continue

			# strip checksum
			if b"*" in data:
				checksum = int(data[data.rfind(b"*") + 1:])
				data = data[:data.rfind(b"*")]
				if not checksum == self._calculate_checksum(data):
					self._triggerResend(expected=self.current_line + 1)
					continue

				self.current_line += 1
			elif settings().getBoolean(["devel", "virtualPrinter", "forceChecksum"]):
				self._send(self._error("checksum_missing"))
				continue

			# track N = N + 1
			if data.startswith(b"N") and b"M110" in data:
				linenumber = int(re.search(b"N([0-9]+)", data).group(1))
				self.lastN = linenumber
				self.current_line = linenumber

				self._triggerResendAt100 = True
				self._triggerResendWithTimeoutAt105 = True

				self._sendOk()
				continue
			elif data.startswith(b"N"):
				linenumber = int(re.search(b"N([0-9]+)", data).group(1))
				expected = self.lastN + 1
				if linenumber != expected:
					self._triggerResend(actual=linenumber)
					continue
				elif linenumber == 100 and self._triggerResendAt100:
					# simulate a resend at line 100
					self._triggerResendAt100 = False
					self._triggerResend(expected=100)
					continue
				elif linenumber == 105 and self._triggerResendWithTimeoutAt105 and not self._writingToSd:
					# simulate a resend with timeout at line 105
					self._triggerResendWithTimeoutAt105 = False
					self._triggerResend(expected=105)
					self._dont_answer = True
					self.lastN = linenumber
					continue
				elif linenumber == 110 and self._triggerResendWithMissingLinenoAt110 and not self._writingToSd:
					self._triggerResendWithMissingLinenoAt110 = False
					self._send(self._error("lineno_missing", self.lastN))
					continue
				elif linenumber == 115 and self._triggerResendWithChecksumMismatchAt115 and not self._writingToSd:
					self._triggerResendWithChecksumMismatchAt115 = False
					self._triggerResend(checksum=True)
					continue
				elif len(self._prepared_errors):
					prepared = self._prepared_errors.pop(0)
					# noinspection PyCompatibility
					if callable(prepared):
						prepared(linenumber, self.lastN, data)
						continue
					elif isinstance(prepared, basestring):
						self._send(prepared)
						continue
				elif self._rerequest_last:
					self._triggerResend(actual=linenumber)
					continue
				else:
					self.lastN = linenumber
				data = data.split(None, 1)[1].strip()

			data += b"\n"

			data = to_unicode(data, encoding="ascii", errors="replace").strip()

			if data.startswith("!!DEBUG:") or data.strip() == "!!DEBUG":
				debug_command = ""
				if data.startswith("!!DEBUG:"):
					debug_command = data[len("!!DEBUG:"):].strip()
				self._debugTrigger(debug_command)
				continue

			# shortcut for writing to SD
			if self._writingToSd and self._writingToSdHandle is not None and not "M29" in data:
				self._writingToSdHandle.write(data)
				self._sendOk()
				continue

			if data.strip() == "version":
				from octoprint import __version__
				self._send("OctoPrint VirtualPrinter v" + __version__)
				continue

			# if we are sending oks before command output, send it now
			if len(data.strip()) > 0 and self._okBeforeCommandOutput:
				self._sendOk()

			# actual command handling
			command_match = VirtualPrinter.command_regex.match(data)
			if command_match is not None:
				command = command_match.group(0)
				letter = command_match.group(1)

				try:
					# if we have a method _gcode_G, _gcode_M or _gcode_T, execute that first
					letter_handler = "_gcode_{}".format(letter)
					if hasattr(self, letter_handler):
						code = command_match.group(2)
						handled = getattr(self, letter_handler)(code, data)
						if handled:
							continue

					# then look for a method _gcode_<command> and execute that if it exists
					command_handler = "_gcode_{}".format(command)
					if hasattr(self, command_handler):
						handled = getattr(self, command_handler)(data)
						if handled:
							continue

				finally:
					# make sure that the debug sleepAfter and sleepAfterNext stuff works even
					# if we continued above
					if len(self._sleepAfter) or len(self._sleepAfterNext):
						interval = None
						if command in self._sleepAfter:
							interval = self._sleepAfter[command]
						elif command in self._sleepAfterNext:
							interval = self._sleepAfterNext[command]
							del self._sleepAfterNext[command]

						if interval is not None:
							self._send("// sleeping for {interval} seconds".format(interval=interval))
							time.sleep(interval)

			# if we are sending oks after command output, send it now
			if len(data.strip()) > 0 and not self._okBeforeCommandOutput:
				self._sendOk()

		self._logger.info("Closing down read loop")

	##~~ command implementations

	# noinspection PyUnusedLocal
	def _gcode_T(self, code, data):
		# type: (str, str) -> None
		t = int(code)
		if 0 <= t < self.extruderCount:
			self.currentExtruder = t
			self._send("Active Extruder: %d" % self.currentExtruder)
		else:
			self._send("echo:T{} Invalid extruder ".format(t))

	# noinspection PyUnusedLocal
	def _gcode_F(self, code, data):
		# type: (str, str) -> bool
		if self._supportF:
			self._send("echo:changed F value")
			return False
		else:
			self._send(self._error("command_unknown", "F"))
			return True

	def _gcode_M104(self, data):
		# type: (str) -> None
		self._parseHotendCommand(data)

	def _gcode_M109(self, data):
		# type: (str) -> None
		self._parseHotendCommand(data, wait=True, support_r=True)

	def _gcode_M140(self, data):
		# type: (str) -> None
		self._parseBedCommand(data)

	def _gcode_M190(self, data):
		# type: (str) -> None
		self._parseBedCommand(data, wait=True, support_r=True)

	def _gcode_M141(self, data):
		self._parseChamberCommand(data)

	def _gcode_M191(self, data):
		self._parseChamberCommand(data, wait=True, support_r=True)

	# noinspection PyUnusedLocal
	def _gcode_M105(self, data):
		# type: (str) -> bool
		self._processTemperatureQuery()
		return True

	# noinspection PyUnusedLocal
	def _gcode_M20(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			self._listSd()

	# noinspection PyUnusedLocal
	def _gcode_M21(self, data):
		# type: (str) -> None
		self._sdCardReady = True
		self._send("SD card ok")

	# noinspection PyUnusedLocal
	def _gcode_M22(self, data):
		# type: (str) -> None
		self._sdCardReady = False

	def _gcode_M23(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			filename = data.split(None, 1)[1].strip()
			self._selectSdFile(filename)

	# noinspection PyUnusedLocal
	def _gcode_M24(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			self._startSdPrint()

	# noinspection PyUnusedLocal
	def _gcode_M25(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			self._pauseSdPrint()

	def _gcode_M26(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			pos = int(re.search(r"S([0-9]+)", data).group(1))
			self._setSdPos(pos)

	def _gcode_M27(self, data):
		# type: (str) -> None
		def report():
			if self._sdCardReady:
				self._reportSdStatus()

		matchS = re.search(r"S([0-9]+)", data)
		if matchS:
			interval = int(matchS.group(1))
			if self._sdstatus_reporter is not None:
				self._sdstatus_reporter.cancel()

			if interval > 0:
				self._sdstatus_reporter = RepeatedTimer(interval, report)
				self._sdstatus_reporter.start()
			else:
				self._sdstatus_reporter = None

		report()

	def _gcode_M28(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			filename = data.split(None, 1)[1].strip()
			self._writeSdFile(filename)

	# noinspection PyUnusedLocal
	def _gcode_M29(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			self._finishSdFile()

	def _gcode_M30(self, data):
		# type: (str) -> None
		if self._sdCardReady:
			filename = data.split(None, 1)[1].strip()
			self._deleteSdFile(filename)

	def _gcode_M113(self, data):
		# type: (str) -> None
		matchS = re.search(r"S([0-9]+)", data)
		if matchS is not None:
			interval = int(matchS.group(1))
			if 0 <= interval <= 60:
				self._busyInterval = interval

	# noinspection PyUnusedLocal
	def _gcode_M114(self, data):
		# type: (str) -> bool
		m114FormatString = settings().get(["devel", "virtualPrinter", "m114FormatString"])
		e = dict((index, value) for index, value in enumerate(self._lastE))
		e["current"] = self._lastE[self.currentExtruder]
		e["all"] = " ".join(["E{}:{}".format(num, self._lastE[self.currentExtruder]) for num in range(self.extruderCount)])
		output = m114FormatString.format(x=self._lastX,
										 y=self._lastY,
										 z=self._lastZ,
										 e=e,
										 f=self._lastF,
										 a=int(self._lastX*100),
										 b=int(self._lastY*100),
										 c=int(self._lastZ*100))

		if not self._okBeforeCommandOutput:
			ok = self._ok()
			if ok:
				output = "{} {}".format(self._ok(), output)
		self._send(output)
		return True

	# noinspection PyUnusedLocal
	def _gcode_M115(self, data):
		# type: (str) -> None
		output = self._m115FormatString.format(firmware_name=self._firmwareName)
		self._send(output)

		if settings().getBoolean(["devel", "virtualPrinter", "m115ReportCapabilities"]):
			for cap, enabled in self._capabilities.items():
				self._send("Cap:{}:{}".format(cap.upper(), "1" if enabled else "0"))

	def _gcode_M117(self, data):
		# type: (str) -> None
		# we'll just use this to echo a message, to allow playing around with pause triggers
		if self._echoOnM117:
			self._send("echo:%s" % re.search(r"M117\s+(.*)", data).group(1))

	def _gcode_M155(self, data):
		# type: (str) -> None
		matchS = re.search(r"S([0-9]+)", data)
		if matchS is not None:
			interval = int(matchS.group(1))
			if self._temperature_reporter is not None:
				self._temperature_reporter.cancel()

			if interval > 0:
				self._temperature_reporter = RepeatedTimer(interval, lambda: self._send(self._generateTemperatureOutput()))
				self._temperature_reporter.start()
			else:
				self._temperature_reporter = None

	def _gcode_M220(self, data):
		# type: (str) -> None
		matchS = re.search(r"S([0-9]+)", data)
		if matchS is not None:
			self._feedrate_multiplier = float(matchS.group(1))

	def _gcode_M221(self, data):
		# type: (str) -> None
		matchS = re.search(r"S([0-9]+)", data)
		if matchS is not None:
			self._flowrate_multiplier = float(matchS.group(1))

	# noinspection PyUnusedLocal
	def _gcode_M400(self, data):
		# type: (str) -> None
		self.buffered.join()

	# noinspection PyUnusedLocal
	def _gcode_M999(self, data):
		# type: (str) -> None
		# mirror Marlin behaviour
		self._send("Resend: 1")

	# noinspection PyUnusedLocal
	def _gcode_G20(self, data):
		# type: (str) -> None
		self._unitModifier = 1.0 / 2.54
		if self._lastX is not None:
			self._lastX *= 2.54
		if self._lastY is not None:
			self._lastY *= 2.54
		if self._lastZ is not None:
			self._lastZ *= 2.54
		if self._lastE is not None:
			self._lastE = [e * 2.54 if e is not None else None for e in self._lastE]

	# noinspection PyUnusedLocal
	def _gcode_G21(self, data):
		# type: (str) -> None
		self._unitModifier = 1.0
		if self._lastX is not None:
			self._lastX /= 2.54
		if self._lastY is not None:
			self._lastY /= 2.54
		if self._lastZ is not None:
			self._lastZ /= 2.54
		if self._lastE is not None:
			self._lastE = [e / 2.54 if e is not None else None for e in self._lastE]

	# noinspection PyUnusedLocal
	def _gcode_G90(self, data):
		# type: (str) -> None
		self._relative = False

	# noinspection PyUnusedLocal
	def _gcode_G91(self, data):
		# type: (str) -> None
		self._relative = True

	def _gcode_G92(self, data):
		# type: (str) -> None
		self._setPosition(data)

	def _gcode_G28(self, data):
		# type: (str) -> None
		self._home(data)

	def _gcode_G0(self, data):
		# type: (str) -> None
		# simulate reprap buffered commands via a Queue with maxsize which internally simulates the moves
		self.buffered.put(data)
	_gcode_G1 = _gcode_G0
	_gcode_G2 = _gcode_G0
	_gcode_G3 = _gcode_G0

	def _gcode_G4(self, data):
		# type: (str) -> None
		matchS = re.search(r'S([0-9]+)', data)
		matchP = re.search(r'P([0-9]+)', data)

		_timeout = 0
		if matchP:
			_timeout = float(matchP.group(1)) / 1000.0
		elif matchS:
			_timeout = float(matchS.group(1))

		if self._sendBusy and self._busyInterval > 0:
			until = monotonic_time() + _timeout
			while monotonic_time() < until:
				time.sleep(self._busyInterval)
				self._send("busy:processing")
		else:
			time.sleep(_timeout)

	# noinspection PyUnusedLocal
	def _gcode_G33(self, data):
		# type: (str) -> None
		self._send("G33 Auto Calibrate")
		self._send("Will take ~60s")
		timeout = 60

		if self._sendBusy and self._busyInterval > 0:
			until = monotonic_time() + timeout
			while monotonic_time() < until:
				time.sleep(self._busyInterval)
				self._send("busy:processing")
		else:
			time.sleep(timeout)

	##~~ further helpers

	# noinspection PyMethodMayBeStatic
	def _calculate_checksum(self, line):
		# type: (bytes) -> int
		checksum = 0
		for c in bytearray(line):
			checksum ^= c
		return checksum

	def _kill(self):
		if not self._supportM112:
			return
		self._killed = True
		self._send("echo:EMERGENCY SHUTDOWN DETECTED. KILLED.")

	def _triggerResend(self, expected=None, actual=None, checksum=None):
		# type: (int, int, int) -> None
		with self._incoming_lock:
			if expected is None:
				expected = self.lastN + 1
			else:
				self.lastN = expected - 1

			if actual is None:
				if checksum:
					self._send(self._error("checksum_mismatch"))
				else:
					self._send(self._error("checksum_missing"))
			else:
				self._send(self._error("lineno_mismatch", expected, actual))

			def request_resend():
				self._send("Resend:%d" % expected)
				if not self._brokenResend:
					self._sendOk()

			request_resend()

	def _debugTrigger(self, data):
		# type: (str) -> None
		if data == "" or data == "help" or data == "?":
			usage = """
			OctoPrint Virtual Printer debug commands

			help
			?
			| This help.

			# Action Triggers

			action_pause
			| Sends a "// action:pause" action trigger to the host.
			action_resume
			| Sends a "// action:resume" action trigger to the host.
			action_disconnect
			| Sends a "// action:disconnect" action trigger to the
			| host.
			action_custom <action>[ <parameters>]
			| Sends a custom "// action:<action> <parameters>"
			| action trigger to the host.

			# Communication Errors

			dont_answer
			| Will not acknowledge the next command.
			go_awol
			| Will completely stop replying
			trigger_resend_lineno
			| Triggers a resend error with a line number mismatch
			trigger_resend_checksum
			| Triggers a resend error with a checksum mismatch
			trigger_missing_checksum
			| Triggers a resend error with a missing checksum
			trigger_missing_lineno
			| Triggers a "no line number with checksum" error w/o resend request
			trigger_fatal_error_marlin
			| Triggers a fatal error/simulated heater fail, Marlin style
			trigger_fatal_error_repetier
			| Triggers a fatal error/simulated heater fail, Repetier style
			drop_connection
			| Drops the serial connection
			prepare_ok <broken ok>
			| Will cause <broken ok> to be enqueued for use,
			| will be used instead of actual "ok"
			rerequest_last
			| Will cause the last line number + 1 to be rerequest add infinitum

			# Reply Timing / Sleeping

			sleep <int:seconds>
			| Sleep <seconds> s
			sleep_after <str:command> <int:seconds>
			| Sleeps <seconds> s after each execution of <command>
			sleep_after_next <str:command> <int:seconds>
			| Sleeps <seconds> s after execution of next <command>

			# SD printing

			start_sd <str:file>
			| Select and start printing file <file> from SD
			select_sd <str:file>
			| Select file <file> from SD, don't start printing it yet. Use
			| start_sd to start the print
			cancel_sd
			| Cancels an ongoing SD print

			# Misc

			send <str:message>
			| Sends back <message>
			reset
			| Simulates a reset. Internal state will be lost.
			"""
			for line in usage.split("\n"):
				self._send("echo: {}".format(line.strip()))
		elif data == "action_pause":
			self._send("// action:pause")
		elif data == "action_resume":
			self._send("// action:resume")
		elif data == "action_disconnect":
			self._send("// action:disconnect")
		elif data == "dont_answer":
			self._dont_answer = True
		elif data == "trigger_resend_lineno":
			self._prepared_errors.append(lambda cur, last, ln: self._triggerResend(expected=last, actual=last+1))
		elif data == "trigger_resend_checksum":
			self._prepared_errors.append(lambda cur, last, ln: self._triggerResend(expected=last, checksum=True))
		elif data == "trigger_missing_checksum":
			self._prepared_errors.append(lambda cur, last, ln: self._triggerResend(expected=last, checksum=False))
		elif data == "trigger_missing_lineno":
			self._prepared_errors.append(lambda cur, last, ln: self._send(self._error("lineno_missing", last)))
		elif data == "trigger_fatal_error_marlin":
			self._send("Error:Thermal Runaway, system stopped! Heater_ID: bed")
			self._send("Error:Printer halted. kill() called!")
		elif data == "trigger_fatal_error_repetier":
			self._send("fatal: Heater/sensor error - Printer stopped and heaters disabled due to this error. Fix error and restart with M999.")
		elif data == "drop_connection":
			self._debug_drop_connection = True
		elif data == "reset":
			self._reset()
		elif data == "mintemp_error":
			self._send(self._error("mintemp"))
		elif data == "maxtemp_error":
			self._send(self._error("maxtemp"))
		elif data == "go_awol":
			self._send("// Going AWOL")
			self._debug_awol = True
		elif data == "rerequest_last":
			self._send("// Entering rerequest loop")
			self._rerequest_last = True
		elif data == "cancel_sd":
			if self._sdPrinting and self._sdPrinter:
				self._pauseSdPrint()
				self._sdPrinting = False
				self._sdPrintingSemaphore.set()
				self._sdPrinter.join()
				self._finishSdPrint()
		else:
			try:
				sleep_match = VirtualPrinter.sleep_regex.match(data)
				sleep_after_match = VirtualPrinter.sleep_after_regex.match(data)
				sleep_after_next_match = VirtualPrinter.sleep_after_next_regex.match(data)
				custom_action_match = VirtualPrinter.custom_action_regex.match(data)
				prepare_ok_match = VirtualPrinter.prepare_ok_regex.match(data)
				send_match = VirtualPrinter.send_regex.match(data)
				set_ambient_match = VirtualPrinter.set_ambient_regex.match(data)
				start_sd_match = VirtualPrinter.start_sd_regex.match(data)
				select_sd_match = VirtualPrinter.select_sd_regex.match(data)

				if sleep_match is not None:
					interval = int(sleep_match.group(1))
					self._send("// sleeping for {interval} seconds".format(interval=interval))
					self._debug_sleep = interval
				elif sleep_after_match is not None:
					command = sleep_after_match.group(1)
					interval = int(sleep_after_match.group(2))
					self._sleepAfter[command] = interval
					self._send("// going to sleep {interval} seconds after each {command}".format(**locals()))
				elif sleep_after_next_match is not None:
					command = sleep_after_next_match.group(1)
					interval = int(sleep_after_next_match.group(2))
					self._sleepAfterNext[command] = interval
					self._send("// going to sleep {interval} seconds after next {command}".format(**locals()))
				elif custom_action_match is not None:
					action = custom_action_match.group(1)
					params = custom_action_match.group(2)
					params = params.strip() if params is not None else ""
					self._send("// action:{action} {params}".format(**locals()).strip())
				elif prepare_ok_match is not None:
					ok = prepare_ok_match.group(1)
					self._prepared_oks.append(ok)
				elif send_match is not None:
					self._send(send_match.group(1))
				elif set_ambient_match is not None:
					self._ambient_temperature = float(set_ambient_match.group(1))
					self._send("// set ambient temperature to {}".format(self._ambient_temperature))
				elif start_sd_match is not None:
					self._selectSdFile(start_sd_match.group(1), check_already_open=True)
					self._startSdPrint()
				elif select_sd_match is not None:
					self._selectSdFile(select_sd_match.group(1))
			except Exception:
				self._logger.exception("While handling %r", data)

	def _listSd(self):
		self._send("Begin file list")
		if settings().getBoolean(["devel", "virtualPrinter", "extendedSdFileList"]):
			items = map(
				lambda x: "%s %d" % (x.upper(), os.stat(os.path.join(self._virtualSd, x)).st_size),
				os.listdir(self._virtualSd)
			)
		else:
			items = map(
				lambda x: x.upper(),
				os.listdir(self._virtualSd)
			)
		for item in items:
			self._send(item)
		self._send("End file list")

	def _selectSdFile(self, filename, check_already_open=False):
		# type: (str, bool) -> None
		if filename.startswith("/"):
			filename = filename[1:]

		file = os.path.join(self._virtualSd, filename.lower())
		if self._selectedSdFile == file and check_already_open:
			return

		if not os.path.exists(file) or not os.path.isfile(file):
			self._send("open failed, File: %s." % filename)
		else:
			self._selectedSdFile = file
			self._selectedSdFileSize = os.stat(file).st_size
			if settings().getBoolean(["devel", "virtualPrinter", "includeFilenameInOpened"]):
				self._send("File opened: %s  Size: %d" % (filename, self._selectedSdFileSize))
			else:
				self._send("File opened")
			self._send("File selected")

	def _startSdPrint(self):
		if self._selectedSdFile is not None:
			if self._sdPrinter is None:
				self._sdPrinting = True
				self._sdPrinter = threading.Thread(target=self._sdPrintingWorker)
				self._sdPrinter.start()
		self._sdPrintingSemaphore.set()

	def _pauseSdPrint(self):
		self._sdPrintingSemaphore.clear()

	def _setSdPos(self, pos):
		self._newSdFilePos = pos

	def _reportSdStatus(self):
		if self._sdPrinter is not None and self._sdPrintingSemaphore.is_set:
			self._send("SD printing byte %d/%d" % (self._selectedSdFilePos, self._selectedSdFileSize))
		else:
			self._send("Not SD printing")

	def _generateTemperatureOutput(self):
		# type: () -> str
		includeTarget = not settings().getBoolean(["devel", "virtualPrinter", "repetierStyleTargetTemperature"])

		# send simulated temperature data
		if self.temperatureCount > 1:
			allTemps = []
			for i in range(len(self.temp)):
				allTemps.append((i, self.temp[i], self.targetTemp[i]))
			allTempsString = " ".join(map(lambda x: "T%d:%.2f /%.2f" % x if includeTarget else "T%d:%.2f" % (x[0], x[1]), allTemps))

			if settings().getBoolean(["devel", "virtualPrinter", "smoothieTemperatureReporting"]):
				allTempsString = allTempsString.replace("T0:", "T:")

			if settings().getBoolean(["devel", "virtualPrinter", "hasBed"]):
				if includeTarget:
					allTempsString = "B:%.2f /%.2f %s" % (self.bedTemp, self.bedTargetTemp, allTempsString)
				else:
					allTempsString = "B:%.2f %s" % (self.bedTemp, allTempsString)

			if settings().getBoolean(["devel", "virtualPrinter", "hasChamber"]):
				if includeTarget:
					allTempsString = "C:%.2f /%.2f %s" % (self.chamberTemp, self.chamberTargetTemp, allTempsString)
				else:
					allTempsString = "C:%.2f %s" % (self.chamberTemp, allTempsString)

			if settings().getBoolean(["devel", "virtualPrinter", "includeCurrentToolInTemps"]):
				if includeTarget:
					output = "T:%.2f /%.2f %s" % (self.temp[self.currentExtruder], self.targetTemp[self.currentExtruder], allTempsString)
				else:
					output = "T:%.2f %s" % (self.temp[self.currentExtruder], allTempsString)
			else:
				output = allTempsString
		else:
			if includeTarget:
				t = "T:%.2f /%.2f" % (self.temp[0], self.targetTemp[0])
			else:
				t = "T:%.2f" % self.temp[0]

			if settings().getBoolean(["devel", "virtualPrinter", "hasBed"]):
				if includeTarget:
					b = "B:%.2f /%.2f" % (self.bedTemp, self.bedTargetTemp)
				else:
					b = "B:%.2f" % self.bedTemp
			else:
				b = ""

			if settings().getBoolean(["devel", "virtualPrinter", "hasChamber"]):
				if includeTarget:
					c = "C:%.2f /%.2f" % (self.chamberTemp, self.chamberTargetTemp)
				else:
					c = "C:%.2f" % self.chamberTemp
			else:
				c = ""

			output = t + " " + b + " " + c
			output = output.strip()

		output += " @:64\n"
		return output

	def _processTemperatureQuery(self):
		includeOk = not self._okBeforeCommandOutput
		output = self._generateTemperatureOutput()

		if includeOk:
			ok = self._ok()
			if ok:
				output = "{} {}".format(ok, output)
		self._send(output)

	def _parseHotendCommand(self, line, wait=False, support_r=False):
		# type: (str, bool, bool) -> None
		only_wait_if_higher = True
		tool = 0
		toolMatch = re.search(r'T([0-9]+)', line)
		if toolMatch:
			tool = int(toolMatch.group(1))

		if tool >= self.temperatureCount:
			return

		try:
			self.targetTemp[tool] = float(re.search(r'S([0-9]+)', line).group(1))
		except Exception:
			if support_r:
				try:
					self.targetTemp[tool] = float(re.search(r'R([0-9]+)', line).group(1))
					only_wait_if_higher = False
				except Exception:
					pass

		if wait:
			self._waitForHeatup("tool%d" % tool, only_wait_if_higher)
		if settings().getBoolean(["devel", "virtualPrinter", "repetierStyleTargetTemperature"]):
			self._send("TargetExtr%d:%d" % (tool, self.targetTemp[tool]))

	def _parseBedCommand(self, line, wait=False, support_r=False):
		# type: (str, bool, bool) -> None
		if not settings().getBoolean(["devel", "virtualPrinter", "hasBed"]):
			return

		only_wait_if_higher = True
		try:
			self.bedTargetTemp = float(re.search(r'S([0-9]+)', line).group(1))
		except Exception:
			if support_r:
				try:
					self.bedTargetTemp = float(re.search(r'R([0-9]+)', line).group(1))
					only_wait_if_higher = False
				except Exception:
					pass

		if wait:
			self._waitForHeatup("bed", only_wait_if_higher)
		if settings().getBoolean(["devel", "virtualPrinter", "repetierStyleTargetTemperature"]):
			self._send("TargetBed:%d" % self.bedTargetTemp)

	def _parseChamberCommand(self, line, wait=False, support_r=False):
		if not settings().getBoolean(["devel", "virtualPrinter", "hasChamber"]):
			return

		only_wait_if_higher = True
		try:
			self.chamberTargetTemp = float(re.search('S([0-9]+)', line).group(1))
		except:
			if support_r:
				try:
					self.chamberTargetTemp = float(re.search('R([0-9]+)', line).group(1))
					only_wait_if_higher = False
				except:
					pass

		if wait:
			self._waitForHeatup("chamber", only_wait_if_higher)

	def _performMove(self, line):
		# type: (str) -> None
		matchX = re.search(r"X(-?[0-9.]+)", line)
		matchY = re.search(r"Y(-?[0-9.]+)", line)
		matchZ = re.search(r"Z(-?[0-9.]+)", line)
		matchE = re.search(r"E(-?[0-9.]+)", line)
		matchF = re.search(r"F([0-9.]+)", line)

		duration = 0.0
		if matchF is not None:
			try:
				self._lastF = float(matchF.group(1))
			except ValueError:
				pass

		speedXYZ = self._lastF * (self._feedrate_multiplier / 100.0)
		speedE = self._lastF * (self._flowrate_multiplier / 100.0)
		if speedXYZ == 0:
			speedXYZ = 999999999999
		if speedE == 0:
			speedE = 999999999999

		if matchX is not None:
			try:
				x = float(matchX.group(1))
			except ValueError:
				pass
			else:
				if self._relative or self._lastX is None:
					duration = max(duration, x * self._unitModifier / speedXYZ * 60.0)
				else:
					duration = max(duration, (x - self._lastX) * self._unitModifier / speedXYZ * 60.0)

				if self._relative and self._lastX is not None:
					self._lastX += x
				else:
					self._lastX = x
		if matchY is not None:
			try:
				y = float(matchY.group(1))
			except ValueError:
				pass
			else:
				if self._relative or self._lastY is None:
					duration = max(duration, y * self._unitModifier / speedXYZ * 60.0)
				else:
					duration = max(duration, (y - self._lastY) * self._unitModifier / speedXYZ * 60.0)

				if self._relative and self._lastY is not None:
					self._lastY += y
				else:
					self._lastY = y
		if matchZ is not None:
			try:
				z = float(matchZ.group(1))
			except ValueError:
				pass
			else:
				if self._relative or self._lastZ is None:
					duration = max(duration, z * self._unitModifier / speedXYZ * 60.0)
				else:
					duration = max(duration, (z - self._lastZ) * self._unitModifier / speedXYZ * 60.0)

				if self._relative and self._lastZ is not None:
					self._lastZ += z
				else:
					self._lastZ = z
		if matchE is not None:
			try:
				e = float(matchE.group(1))
			except ValueError:
				pass
			else:
				lastE = self._lastE[self.currentExtruder]
				if self._relative or lastE is None:
					duration = max(duration, e * self._unitModifier / speedE * 60.0)
				else:
					duration = max(duration, (e - lastE) * self._unitModifier / speedE * 60.0)

				if self._relative and lastE is not None:
					self._lastE[self.currentExtruder] += e
				else:
					self._lastE[self.currentExtruder] = e

		if duration:
			duration *= 0.1
			if duration > self._read_timeout:
				slept = 0
				while duration - slept > self._read_timeout and not self._killed:
					time.sleep(self._read_timeout)
					slept += self._read_timeout
			else:
				time.sleep(duration)

	def _setPosition(self, line):
		# type: (str) -> None
		matchX = re.search(r"X(-?[0-9.]+)", line)
		matchY = re.search(r"Y(-?[0-9.]+)", line)
		matchZ = re.search(r"Z(-?[0-9.]+)", line)
		matchE = re.search(r"E(-?[0-9.]+)", line)

		if matchX is None and matchY is None and matchZ is None and matchE is None:
			self._lastX = self._lastY = self._lastZ = self._lastE[self.currentExtruder] = 0
		else:
			if matchX is not None:
				try:
					self._lastX = float(matchX.group(1))
				except ValueError:
					pass
			if matchY is not None:
				try:
					self._lastY = float(matchY.group(1))
				except ValueError:
					pass
			if matchZ is not None:
				try:
					self._lastZ = float(matchZ.group(1))
				except ValueError:
					pass
			if matchE is not None:
				try:
					self._lastE[self.currentExtruder] = float(matchE.group(1))
				except ValueError:
					pass

	def _home(self, line):
		x = y = z = e = None

		if "X" in line:
			x = True
		if "Y" in line:
			y = True
		if "Z" in line:
			z = True
		if "E" in line:
			e = True

		if x is None and y is None and z is None and e is None:
			self._lastX = self._lastY = self._lastZ = self._lastE[self.currentExtruder] = 0
		else:
			if x:
				self._lastX = 0
			if y:
				self._lastY = 0
			if z:
				self._lastZ = 0
			if e:
				self._lastE = 0

	def _writeSdFile(self, filename):
		# type: (str) -> None
		filename = filename
		if filename.startswith("/"):
			filename = filename[1:]
		file = os.path.join(self._virtualSd, filename).lower()
		if os.path.exists(file):
			if os.path.isfile(file):
				os.remove(file)
			else:
				self._send("error writing to file")

		handle = None
		try:
			handle = io.open(file, 'wt', encoding='utf-8')
		except Exception:
			self._send("error writing to file")
		self._writingToSdHandle = handle
		self._writingToSd = True
		self._selectedSdFile = file
		self._send("Writing to file: %s" % filename)

	def _finishSdFile(self):
		try:
			self._writingToSdHandle.close()
		except Exception:
			pass
		finally:
			self._writingToSdHandle = None
		self._writingToSd = False
		self._selectedSdFile = None
		self._send("Done saving file")

	def _sdPrintingWorker(self):
		self._selectedSdFilePos = 0
		try:
			with io.open(self._selectedSdFile, 'rt', encoding='utf-8') as f:
				for line in iter(f.readline, ""):
					if self._killed or not self._sdPrinting:
						break

					# reset position if requested by client
					if self._newSdFilePos is not None:
						f.seek(self._newSdFilePos)
						self._newSdFilePos = None

					# read current file position
					self._selectedSdFilePos = f.tell()

					# if we are paused, wait for resuming
					self._sdPrintingSemaphore.wait()
					if self._killed or not self._sdPrinting:
						break

					# set target temps
					if 'M104' in line or 'M109' in line:
						self._parseHotendCommand(line, wait='M109' in line)
					elif 'M140' in line or 'M190' in line:
						self._parseBedCommand(line, wait='M190' in line)
					elif line.startswith("G0") or line.startswith("G1") or line.startswith("G2") or line.startswith("G3"):
						# simulate reprap buffered commands via a Queue with maxsize which internally simulates the moves
						self.buffered.put(line)

		except AttributeError:
			if self.outgoing is not None:
				raise

		self._finishSdPrint()

	def _finishSdPrint(self):
		if not self._killed:
			self._sdPrintingSemaphore.clear()
			self._send("Done printing file")
			self._selectedSdFilePos = 0
			self._sdPrinting = False
			self._sdPrinter = None

	def _waitForHeatup(self, heater, only_wait_if_higher):
		# type: (str, bool) -> None
		delta = 1
		delay = 1
		last_busy = monotonic_time()

		self._heatingUp = True
		try:
			if heater.startswith("tool"):
				toolNum = int(heater[len("tool"):])
				test = lambda: self.temp[toolNum] < self.targetTemp[toolNum] - delta or (not only_wait_if_higher and self.temp[toolNum] > self.targetTemp[toolNum] + delta)
				output = lambda: "T:%0.2f" % self.temp[toolNum]
			elif heater == "bed":
				test = lambda: self.bedTemp < self.bedTargetTemp - delta or (not only_wait_if_higher and self.bedTemp > self.bedTargetTemp + delta)
				output = lambda: "B:%0.2f" % self.bedTemp
			elif heater == "chamber":
				test = lambda: self.chamberTemp < self.chamberTargetTemp - delta or (not only_wait_if_higher and self.chamberTemp > self.chamberTargetTemp + delta)
				output = lambda: "C:%0.2f" % self.chamberTemp
			else:
				return

			while not self._killed and self._heatingUp and test():
				self._simulateTemps(delta=delta)
				self._send(output())
				if self._sendBusy and monotonic_time() - last_busy >= self._busyInterval:
					self._send("echo:busy: processing")
					last_busy = monotonic_time()
				time.sleep(delay)
		except AttributeError:
			if self.outgoing is not None:
				raise
		finally:
			self._heatingUp = False

	def _deleteSdFile(self, filename):
		# type: (str) -> None
		if filename.startswith("/"):
			filename = filename[1:]
		f = os.path.join(self._virtualSd, filename)
		if os.path.exists(f) and os.path.isfile(f):
			os.remove(f)

	def _simulateTemps(self, delta=0.5):
		timeDiff = self.lastTempAt - monotonic_time()
		self.lastTempAt = monotonic_time()

		def simulate(actual, target, ambient):
			if target > 0 and abs(actual - target) > delta:
				goal = target
				factor = 10
			elif not target and abs(actual - ambient) > delta:
				goal = ambient
				factor = 2
			else:
				return actual

			old = actual
			actual += math.copysign(timeDiff * factor, goal - actual)

			if math.copysign(1, goal - old) != math.copysign(1, goal - actual):
				actual = goal

			return actual

		for i in range(len(self.temp)):
			if i in self.pinnedExtruders:
				self.temp[i] = self.pinnedExtruders[i]
				continue
			self.temp[i] = simulate(self.temp[i], self.targetTemp[i], self._ambient_temperature)
		self.bedTemp = simulate(self.bedTemp, self.bedTargetTemp, self._ambient_temperature)
		self.chamberTemp = simulate(self.chamberTemp, self.chamberTargetTemp, self._ambient_temperature)

	def _processBuffer(self):
		while self.buffered is not None:
			try:
				line = self.buffered.get(timeout=0.5)
			except queue.Empty:
				continue

			if line is None:
				continue

			self._performMove(line)
			self.buffered.task_done()

		self._logger.info("Closing down buffer loop")

	def write(self, data):
		# type: (bytes) -> int
		data = to_bytes(data, errors="replace")
		u_data = to_unicode(data, errors="replace")

		if self._debug_awol:
			return len(data)

		if self._debug_drop_connection:
			self._logger.info("Debug drop of connection requested, raising SerialTimeoutException")
			raise SerialTimeoutException()

		with self._incoming_lock:
			if self.incoming is None or self.outgoing is None:
				return 0

			if b"M112" in data and self._supportM112:
				self._seriallog.info("<<< {}".format(u_data))
				self._kill()
				return len(data)

			try:
				written = self.incoming.put(data, timeout=self._write_timeout, partial=True)
				self._seriallog.info("<<< {}".format(u_data))
				return written
			except queue.Full:
				self._logger.info("Incoming queue is full, raising SerialTimeoutException")
				raise SerialTimeoutException()

	def readline(self):
		# type: () -> bytes
		if self._debug_awol:
			time.sleep(self._read_timeout)
			return b""

		if self._debug_drop_connection:
			raise SerialTimeoutException()

		if self._debug_sleep > 0:
			# if we are supposed to sleep, we sleep not longer than the read timeout
			# (and then on the next call sleep again if there's time to sleep left)
			sleep_for = min(self._debug_sleep, self._read_timeout)
			self._debug_sleep -= sleep_for
			time.sleep(sleep_for)

			if self._debug_sleep > 0:
				# we slept the full read timeout, return an empty line
				return b""

			# otherwise our left over timeout is the read timeout minus what we already
			# slept for
			timeout = self._read_timeout - sleep_for

		else:
			# use the full read timeout as timeout
			timeout = self._read_timeout

		try:
			# fetch a line from the queue, wait no longer than timeout
			line = to_unicode(self.outgoing.get(timeout=timeout), errors="replace")
			self._seriallog.info(u">>> {}".format(line.strip()))
			self.outgoing.task_done()
			return to_bytes(line)
		except queue.Empty:
			# queue empty? return empty line
			return b""

	def close(self):
		self._killed = True
		self.incoming = None
		self.outgoing = None
		self.buffered = None

	def _sendOk(self):
		if self.outgoing is None:
			return
		ok = self._ok()
		if ok:
			self._send(ok)

	def _sendWaitAfterTimeout(self, timeout=5):
		time.sleep(timeout)
		if self.outgoing is not None:
			self._send("wait")

	def _send(self, line):
		# type: (str) -> None
		if self.outgoing is not None:
			self.outgoing.put(line)

	def _ok(self):
		ok = self._okFormatString
		if self._prepared_oks:
			ok = self._prepared_oks.pop(0)
			if ok is None:
				return ok

		return ok.format(ok, lastN=self.lastN, buffer=self.buffered.maxsize - self.buffered.qsize())

	def _error(self, error, *args, **kwargs):
		# type: (str, Any, Any) -> str
		return "Error: {}".format(self._errors.get(error).format(*args, **kwargs))


# noinspection PyUnresolvedReferences
class CharCountingQueue(queue.Queue):

	def __init__(self, maxsize, name=None):
		queue.Queue.__init__(self, maxsize=maxsize)
		self._size = 0
		self._name = name

	def clear(self):
		with self.mutex:
			self.queue.clear()

	def put(self, item, block=True, timeout=None, partial=False):
		self.not_full.acquire()

		try:
			if not self._will_it_fit(item) and partial:
				space_left = self.maxsize - self._qsize()
				if space_left:
					item = item[:space_left]

			if not block:
				if not self._will_it_fit(item):
					raise queue.Full
			elif timeout is None:
				while not self._will_it_fit(item):
					self.not_full.wait()
			elif timeout < 0:
				raise ValueError("'timeout' must be a positive number")
			else:
				endtime = monotonic_time() + timeout
				while not self._will_it_fit(item):
					remaining = endtime - monotonic_time()
					if remaining <= 0.0:
						raise queue.Full
					self.not_full.wait(remaining)

			self._put(item)
			self.unfinished_tasks += 1
			self.not_empty.notify()

			return self._len(item)
		finally:
			self.not_full.release()

	# noinspection PyMethodMayBeStatic
	def _len(self, item):
		return len(item)

	def _qsize(self, l=len):
		return self._size

	# Put a new item in the queue
	def _put(self, item):
		self.queue.append(item)
		self._size += self._len(item)

	# Get an item from the queue
	def _get(self):
		item = self.queue.popleft()
		self._size -= self._len(item)
		return item

	def _will_it_fit(self, item):
		return self.maxsize - self._qsize() >= self._len(item)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin

from octoprint.settings import default_settings
from octoprint.plugin.core import FolderOrigin
from octoprint.server import admin_permission, NO_CONTENT
from octoprint.server.util.flask import no_firstrun_access
from octoprint.util import is_hidden_path, to_bytes
from octoprint.util.version import get_octoprint_version_string, get_octoprint_version, get_comparable_version, is_octoprint_compatible
from octoprint.util.platform import is_os_compatible
from octoprint.util.pip import LocalPipCaller
from octoprint.access import ADMIN_GROUP
from octoprint.access.permissions import Permissions

try:
	from os import scandir
except ImportError:
	from scandir import scandir

try:
	import zlib # check if zlib is available
except ImportError:
	zlib = None


from flask_babel import gettext

import io
import flask
import logging
import os
import requests
import sarge
import shutil
import tempfile
import threading
import time
import zipfile
import json
import sys
import traceback


UNKNOWN_PLUGINS_FILE = "unknown_plugins_from_restore.json"

BACKUP_FILE_PREFIX = "octoprint-backup"

BACKUP_DATE_TIME_FMT = "%Y%m%d-%H%M%S"

def build_backup_filename():
	return "{}-{}.zip".format(BACKUP_FILE_PREFIX,
							  time.strftime(BACKUP_DATE_TIME_FMT))


class BackupPlugin(octoprint.plugin.SettingsPlugin,
                   octoprint.plugin.TemplatePlugin,
                   octoprint.plugin.AssetPlugin,
                   octoprint.plugin.BlueprintPlugin,
                   octoprint.plugin.StartupPlugin):

	_pip_caller = None

	# noinspection PyMissingConstructor
	def __init__(self):
		self._in_progress = []
		self._in_progress_lock = threading.RLock()

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="ACCESS",
			     name="Backup access",
			     description=gettext("Allows access to backups and restores"),
			     roles=["access"],
			     dangerous=True,
			     default_groups=[ADMIN_GROUP])
		]

	##~~ StartupPlugin

	def on_after_startup(self):
		self._clean_dir_backup(self._settings._basedir,
		                       on_log_progress=self._logger.info)

	##~~ AssetPlugin

	def get_assets(self):
		return dict(js=["js/backup.js"],
		            clientjs=["clientjs/backup.js"],
		            css=["css/backup.css"],
		            less=["less/backup.less"])

	##~~ TemplatePlugin

	def get_template_configs(self):
		return [dict(type="settings", name=gettext("Backup & Restore"))]

	##~~ BlueprintPlugin

	@octoprint.plugin.BlueprintPlugin.route("/", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_BACKUP_ACCESS.require(403)
	def get_state(self):
		backups = self._get_backups()
		unknown_plugins = self._get_unknown_plugins()
		return flask.jsonify(backups=backups,
		                     backup_in_progress=len(self._in_progress) > 0,
		                     unknown_plugins=unknown_plugins,
		                     restore_supported=is_os_compatible(["!windows"]))

	@octoprint.plugin.BlueprintPlugin.route("/unknown_plugins", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_BACKUP_ACCESS.require(403)
	def get_unknown_plugins(self):
		# TODO add caching
		unknown_plugins = self._get_unknown_plugins()
		return flask.jsonify(unknown_plugins=unknown_plugins)

	@octoprint.plugin.BlueprintPlugin.route("/unknown_plugins", methods=["DELETE"])
	@no_firstrun_access
	@Permissions.PLUGIN_BACKUP_ACCESS.require(403)
	def delete_unknown_plugins(self):
		data_file = os.path.join(self.get_plugin_data_folder(), UNKNOWN_PLUGINS_FILE)
		try:
			os.remove(data_file)
		except Exception:
			pass
		return NO_CONTENT

	@octoprint.plugin.BlueprintPlugin.route("/backup", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_BACKUP_ACCESS.require(403)
	def get_backups(self):
		backups = self._get_backups()
		return flask.jsonify(backups=backups)

	@octoprint.plugin.BlueprintPlugin.route("/backup", methods=["POST"])
	@no_firstrun_access
	@Permissions.PLUGIN_BACKUP_ACCESS.require(403)
	def create_backup(self):
		backup_file = build_backup_filename()

		data = flask.request.json
		exclude = data.get("exclude", [])

		def on_backup_start(name, temporary_path, exclude):
			self._logger.info("Creating backup zip at {} (excluded: {})...".format(temporary_path,
			                                                                       ",".join(exclude) if len(exclude) else "-"))

			with self._in_progress_lock:
				self._in_progress.append(name)
				self._send_client_message("backup_started", payload=dict(name=name))

		def on_backup_done(name, final_path, exclude):
			with self._in_progress_lock:
				self._in_progress.remove(name)
				self._send_client_message("backup_done", payload=dict(name=name))

			self._logger.info("... done creating backup zip.")

		def on_backup_error(name, exc_info):
			with self._in_progress_lock:
				try:
					self._in_progress.remove(name)
				except ValueError:
					# we'll ignore that
					pass

			self._send_client_message("backup_error", payload=dict(name=name,
			                                                       error="{}".format(exc_info[1])))
			self._logger.error("Error while creating backup zip", exc_info=exc_info)

		thread = threading.Thread(target=self._create_backup,
		                          args=(backup_file,),
		                          kwargs=dict(exclude=exclude,
		                                      settings=self._settings,
		                                      plugin_manager=self._plugin_manager,
		                                      datafolder=self.get_plugin_data_folder(),
		                                      on_backup_start=on_backup_start,
		                                      on_backup_done=on_backup_done,
		                                      on_backup_error=on_backup_error))
		thread.daemon = True
		thread.start()

		response = flask.jsonify(started=True,
		                         name=backup_file)
		response.status_code = 201
		return response

	@octoprint.plugin.BlueprintPlugin.route("/backup/<filename>", methods=["DELETE"])
	@no_firstrun_access
	@Permissions.PLUGIN_BACKUP_ACCESS.require(403)
	def delete_backup(self, filename):
		backup_folder = self.get_plugin_data_folder()
		full_path = os.path.realpath(os.path.join(backup_folder, filename))
		if full_path.startswith(backup_folder) \
			and os.path.exists(full_path) \
			and not is_hidden_path(full_path):
			try:
				os.remove(full_path)
			except Exception:
				self._logger.exception("Could not delete {}".format(filename))
				raise
		return NO_CONTENT

	@octoprint.plugin.BlueprintPlugin.route("/restore", methods=["POST"])
	@no_firstrun_access
	@Permissions.PLUGIN_BACKUP_ACCESS.require(403)
	def perform_restore(self):
		if not is_os_compatible(["!windows"]):
			return flask.make_response("Invalid request, the restores are not supported on the underlying operating system", 400)

		input_name = "file"
		input_upload_path = input_name + "." + self._settings.global_get(["server", "uploads", "pathSuffix"])

		if input_upload_path in flask.request.values:
			# file to restore was uploaded
			path = flask.request.values[input_upload_path]

		elif flask.request.json and "path" in flask.request.json:
			# existing backup is supposed to be restored
			backup_folder = self.get_plugin_data_folder()
			path = os.path.realpath(os.path.join(backup_folder, flask.request.json["path"]))
			if not path.startswith(backup_folder) \
				or not os.path.exists(path) \
				or is_hidden_path(path):
				return flask.abort(404)

		else:
			return flask.make_response("Invalid request, neither a file nor a path of a file to restore provided", 400)

		def on_install_plugins(plugins):
			force_user = self._settings.global_get_boolean(["plugins", "pluginmanager", "pip_force_user"])
			pip_args = self._settings.global_get(["plugins", "pluginmanager", "pip_args"])

			def on_log(line):
				self._logger.info(line)
				self._send_client_message("logline", dict(line=line, type="stdout"))

			for plugin in plugins:
				octoprint_compatible = is_octoprint_compatible(*plugin["compatibility"]["octoprint"])
				os_compatible = is_os_compatible(plugin["compatibility"]["os"])
				compatible = octoprint_compatible and os_compatible
				if not compatible:
					if not octoprint_compatible and not os_compatible:
						self._logger.warn("Cannot install plugin {}, it is incompatible to this version "
						                  "of OctoPrint and the underlying operating system".format(plugin["id"]))
					elif not octoprint_compatible:
						self._logger.warn("Cannot install plugin {}, it is incompatible to this version "
						                  "of OctoPrint".format(plugin["id"]))
					elif not os_compatible:
						self._logger.warn("Cannot install plugin {}, it is incompatible to the underlying "
						                  "operating system".format(plugin["id"]))
					self._send_client_message("plugin_incompatible", dict(plugin=plugin["id"],
					                                                      octoprint_compatible=octoprint_compatible,
					                                                      os_compatible=os_compatible))
					continue

				self._logger.info("Installing plugin {}".format(plugin["id"]))
				self._send_client_message("installing_plugin", dict(plugin=plugin["id"]))
				self.__class__._install_plugin(plugin,
				                               force_user=force_user,
				                               pip_args=pip_args,
				                               on_log=on_log)

		def on_report_unknown_plugins(plugins):
			self._send_client_message("unknown_plugins", payload=dict(plugins=plugins))

		def on_log_progress(line):
			self._logger.info(line)
			self._send_client_message("logline", payload=dict(line=line, stream="stdout"))

		def on_log_error(line, exc_info=None):
			self._logger.error(line, exc_info=exc_info)
			self._send_client_message("logline", payload=dict(line=line, stream="stderr"))

			if exc_info is not None:
				exc_type, exc_value, exc_tb = exc_info
				output = traceback.format_exception(exc_type, exc_value, exc_tb)
				for line in output:
					self._send_client_message("logline", payload=dict(line=line.rstrip(), stream="stderr"))

		def on_restore_start(path):
			self._send_client_message("restore_started")

		def on_restore_done(path):
			self._send_client_message("restore_done")

		def on_restore_failed(path):
			self._send_client_message("restore_failed")

		def on_invalid_backup(line):
			on_log_error(line)

		archive = tempfile.NamedTemporaryFile(delete=False)
		archive.close()
		shutil.copy(path, archive.name)
		path = archive.name

		# noinspection PyTypeChecker
		thread = threading.Thread(target=self._restore_backup,
		                          args=(path,),
		                          kwargs=dict(settings=self._settings,
		                                      plugin_manager=self._plugin_manager,
		                                      datafolder=self.get_plugin_data_folder(),
		                                      on_install_plugins=on_install_plugins,
		                                      on_report_unknown_plugins=on_report_unknown_plugins,
		                                      on_invalid_backup=on_invalid_backup,
		                                      on_log_progress=on_log_progress,
		                                      on_log_error=on_log_error,
		                                      on_restore_start=on_restore_start,
		                                      on_restore_done=on_restore_done,
		                                      on_restore_failed=on_restore_failed))
		thread.daemon = True
		thread.start()

		return flask.jsonify(started=True)

	def is_blueprint_protected(self):
		return False

	##~~ tornado hook

	def route_hook(self, *args, **kwargs):
		from octoprint.server.util.tornado import LargeResponseHandler, path_validation_factory
		from octoprint.util import is_hidden_path
		from octoprint.server import app
		from octoprint.server.util.tornado import access_validation_factory
		from octoprint.server.util.flask import admin_validator

		return [
			(r"/download/(.*)", LargeResponseHandler, dict(path=self.get_plugin_data_folder(),
			                                               as_attachment=True,
			                                               path_validation=path_validation_factory(lambda path: not is_hidden_path(path),
			                                                                                       status_code=404),
			                                               access_validation=access_validation_factory(app, admin_validator)))
		]

	def bodysize_hook(self, current_max_body_sizes, *args, **kwargs):
		# max upload size of 1GB for the restore endpoint
		return [("POST", r"/restore", 1024 * 1024 * 1024)]

	##~~ CLI hook

	def cli_commands_hook(self, cli_group, pass_octoprint_ctx, *args, **kwargs):
		import click

		@click.command("backup")
		@click.option("--exclude", multiple=True,
		              help="Identifiers of data folders to exclude, e.g. 'uploads' to exclude uploads or "
		                   "'timelapse' to exclude timelapses.")
		def backup_command(exclude):
			"""
			Creates a new backup.
			"""

			backup_file = build_backup_filename()
			settings = octoprint.plugin.plugin_settings_for_settings_plugin("backup", self, settings=cli_group.settings)

			datafolder = os.path.join(settings.getBaseFolder("data"), "backup")
			if not os.path.isdir(datafolder):
				os.makedirs(datafolder)

			click.echo("Creating backup at {}, please wait...".format(backup_file))
			self._create_backup(backup_file,
			                    exclude=exclude,
			                    settings=settings,
			                    plugin_manager=cli_group.plugin_manager,
			                    datafolder=datafolder)
			click.echo("Done.")

		@click.command("restore")
		@click.argument("path")
		def restore_command(path):
			"""
			Restores an existing backup from the backup zip provided as argument.

			OctoPrint does not need to run for this to proceed.
			"""
			settings = octoprint.plugin.plugin_settings_for_settings_plugin("backup", self, settings=cli_group.settings)
			plugin_manager = cli_group.plugin_manager

			datafolder = os.path.join(settings.getBaseFolder("data"), "backup")
			if not os.path.isdir(datafolder):
				os.makedirs(datafolder)

			# register plugin manager plugin setting overlays
			plugin_info = plugin_manager.get_plugin_info("pluginmanager")
			if plugin_info and plugin_info.implementation:
				default_settings_overlay = dict(plugins=dict())
				default_settings_overlay["plugins"]["pluginmanager"] = plugin_info.implementation.get_settings_defaults()
				settings.add_overlay(default_settings_overlay, at_end=True)

			if not os.path.isabs(path):
				datafolder = os.path.join(settings.getBaseFolder("data"), "backup")
				if not os.path.isdir(datafolder):
					os.makedirs(datafolder)
				path = os.path.join(datafolder, path)

			if not os.path.exists(path):
				click.echo("Backup {} does not exist".format(path), err=True)
				sys.exit(-1)

			archive = tempfile.NamedTemporaryFile(delete=False)
			archive.close()
			shutil.copy(path, archive.name)
			path = archive.name

			def on_install_plugins(plugins):
				if not plugins:
					return

				force_user = settings.global_get_boolean(["plugins", "pluginmanager", "pip_force_user"])
				pip_args = settings.global_get(["plugins", "pluginmanager", "pip_args"])

				def log(line):
					click.echo("\t{}".format(line))

				for plugin in plugins:
					octoprint_compatible = is_octoprint_compatible(*plugin["compatibility"]["octoprint"])
					os_compatible = is_os_compatible(plugin["compatibility"]["os"])
					compatible = octoprint_compatible and os_compatible
					if not compatible:
						if not octoprint_compatible and not os_compatible:
							click.echo("Cannot install plugin {}, it is incompatible to this version of "
							           "OctoPrint and the underlying operating system".format(plugin["id"]))
						elif not octoprint_compatible:
							click.echo("Cannot install plugin {}, it is incompatible to this version of "
							           "OctoPrint".format(plugin["id"]))
						elif not os_compatible:
							click.echo("Cannot install plugin {}, it is incompatible to the underlying "
							           "operating system".format(plugin["id"]))
						continue

					click.echo("Installing plugin {}".format(plugin["id"]))
					self.__class__._install_plugin(plugin,
					                               force_user=force_user,
					                               pip_args=pip_args,
					                               on_log=log)

			def on_report_unknown_plugins(plugins):
				if not plugins:
					return

				click.echo("The following plugins were not found in the plugin repository. You'll need to install them manually.")
				for plugin in plugins:
					click.echo("\t{} (Homepage: {})".format(plugin["name"], plugin["url"] if plugin["url"] else "?"))

			def on_log_progress(line):
				click.echo(line)

			def on_log_error(line, exc_info=None):
				click.echo(line, err=True)

				if exc_info is not None:
					exc_type, exc_value, exc_tb = exc_info
					output = traceback.format_exception(exc_type, exc_value, exc_tb)
					for line in output:
						click.echo(line.rstrip(), err=True)

			if self._restore_backup(path,
			                        settings=settings,
			                        plugin_manager=plugin_manager,
			                        datafolder=datafolder,
			                        on_install_plugins=on_install_plugins,
			                        on_report_unknown_plugins=on_report_unknown_plugins,
			                        on_log_progress=on_log_progress,
			                        on_log_error=on_log_error,
			                        on_invalid_backup=on_log_error):
				click.echo("Restored from {}".format(path))
			else:
				click.echo("Restoring from {} failed".format(path), err=True)

		return [backup_command, restore_command]

	##~~ helpers

	def _get_backups(self):
		backups = []
		for entry in scandir(self.get_plugin_data_folder()):
			if is_hidden_path(entry.path):
				continue
			if not entry.is_file():
				continue
			if not entry.name.endswith(".zip"):
				continue

			backups.append(dict(name=entry.name,
			                    date=entry.stat().st_mtime,
			                    size=entry.stat().st_size,
			                    url=flask.url_for("index") + "plugin/backup/download/" + entry.name))
		return backups

	def _get_unknown_plugins(self):
		data_file = os.path.join(self.get_plugin_data_folder(), UNKNOWN_PLUGINS_FILE)
		if os.path.exists(data_file):
			try:
				with io.open(data_file, mode='rb', encoding="utf-8") as f:
					unknown_plugins = json.load(f)

				assert isinstance(unknown_plugins, list)
				assert all(map(lambda x: isinstance(x, dict) and "key" in x and "name" in x and "url" in x, unknown_plugins))

				installed_plugins = self._plugin_manager.plugins
				unknown_plugins = list(filter(lambda x: x["key"] not in installed_plugins, unknown_plugins))
				if not unknown_plugins:
					# no plugins left uninstalled, delete data file
					try:
						os.remove(data_file)
					except Exception:
						self._logger.exception("Error while deleting list of unknown plugins at {}".format(data_file))

				return unknown_plugins
			except Exception:
				self._logger.exception("Error while reading list of unknown plugins from {}".format(data_file))
				try:
					os.remove(data_file)
				except Exception:
					self._logger.exception("Error while deleting list of unknown plugins at {}".format(data_file))

		return []

	@classmethod
	def _clean_dir_backup(cls, basedir,
	                      on_log_progress=None):
		basedir_backup = basedir + ".bck"

		if os.path.exists(basedir_backup):
			def remove_bck():
				if callable(on_log_progress):
					on_log_progress("Found config folder backup from prior restore, deleting it...")
				shutil.rmtree(basedir_backup)
				if callable(on_log_progress):
					on_log_progress("... deleted.")

			thread = threading.Thread(target=remove_bck)
			thread.daemon = True
			thread.start()

	@classmethod
	def _get_disk_size(cls, path, ignored=None):
		if ignored is None:
			ignored = []

		if path in ignored:
			return 0

		total = 0
		for entry in scandir(path):
			if entry.is_dir():
				total += cls._get_disk_size(entry.path, ignored=ignored)
			elif entry.is_file():
				total += entry.stat().st_size
		return total

	@classmethod
	def _free_space(cls, path, size):
		from psutil import disk_usage
		return disk_usage(path).free > size

	@classmethod
	def _get_plugin_repository_data(cls, url, logger=None):
		if logger is None:
			logger = logging.getLogger(__name__)

		try:
			r = requests.get(url, timeout=30)
			r.raise_for_status()
		except Exception:
			logger.exception("Error while fetching the plugin repository data from {}".format(url))
			return dict()

		from octoprint.plugins.pluginmanager import map_repository_entry
		return dict((plugin["id"], plugin) for plugin in map(map_repository_entry, r.json()))

	@classmethod
	def _install_plugin(cls, plugin, force_user=False, pip_args=None, on_log=None):
		if pip_args is None:
			pip_args = []

		if on_log is None:
			on_log = logging.getLogger(__name__).info

		# prepare pip caller
		def log(prefix, *lines):
			for line in lines:
				on_log("{} {}".format(prefix, line.rstrip()))

		def log_call(*lines):
			log(">", *lines)

		def log_stdout(*lines):
			log("<", *lines)

		def log_stderr(*lines):
			log("!", *lines)

		if cls._pip_caller is None:
			cls._pip_caller = LocalPipCaller(force_user=force_user)

		cls._pip_caller.on_log_call = log_call
		cls._pip_caller.on_log_stdout = log_stdout
		cls._pip_caller.on_log_stderr = log_stderr

		# install plugin
		pip = ["install", sarge.shell_quote(plugin["archive"]), '--no-cache-dir']

		if plugin.get("follow_dependency_links"):
			pip.append("--process-dependency-links")

		if force_user:
			pip.append("--user")

		if pip_args:
			pip += pip_args

		cls._pip_caller.execute(*pip)

	@classmethod
	def _create_backup(cls, name,
	                   exclude=None,
	                   settings=None,
	                   plugin_manager=None,
	                   datafolder=None,
	                   on_backup_start=None,
	                   on_backup_done=None,
	                   on_backup_error=None):
		exclude_by_default = ("generated", "logs", "watched",)

		try:
			if exclude is None:
				exclude = []
			if not isinstance(exclude, list):
				exclude = list(exclude)

			if "timelapse" in exclude:
				exclude.append("timelapse_tmp")

			configfile = settings._configfile
			basedir = settings._basedir

			temporary_path = os.path.join(datafolder, ".{}".format(name))
			final_path = os.path.join(datafolder, name)

			own_folder = datafolder
			defaults = [os.path.join(basedir, "config.yaml"),] + \
			           [os.path.join(basedir, folder) for folder in default_settings["folder"].keys()]

			# check how much many bytes we are about to backup
			size = os.stat(configfile).st_size
			for folder in default_settings["folder"].keys():
				if folder in exclude or folder in exclude_by_default:
					continue
				size += cls._get_disk_size(settings.global_get_basefolder(folder),
				                           ignored=[own_folder,])
			size += cls._get_disk_size(basedir, ignored=defaults + [own_folder,])

			# since we can't know the compression ratio beforehand, we assume we need the same amount of space
			if not cls._free_space(os.path.dirname(temporary_path), size):
				raise InsufficientSpace()

			compression = zipfile.ZIP_DEFLATED if zlib else zipfile.ZIP_STORED

			if callable(on_backup_start):
				on_backup_start(name, temporary_path, exclude)

			with zipfile.ZipFile(temporary_path, mode="w", compression=compression, allowZip64=True) as zip:
				def add_to_zip(source, target, ignored=None):
					if ignored is None:
						ignored = []

					if source in ignored:
						return

					if os.path.isdir(source):
						for entry in scandir(source):
							add_to_zip(entry.path, os.path.join(target, entry.name), ignored=ignored)
					elif os.path.isfile(source):
						zip.write(source, arcname=target)

				# add metadata
				metadata = dict(version=get_octoprint_version_string(),
				                excludes=exclude)
				zip.writestr("metadata.json", json.dumps(metadata))

				# backup current config file
				add_to_zip(configfile, "basedir/config.yaml", ignored=[own_folder,])

				# backup configured folder paths
				for folder in default_settings["folder"].keys():
					if folder in exclude or folder in exclude_by_default:
						continue

					add_to_zip(settings.global_get_basefolder(folder),
					           "basedir/" + folder.replace("_", "/"),
					           ignored=[own_folder,])

				# backup anything else that might be lying around in our basedir
				add_to_zip(basedir, "basedir", ignored=defaults + [own_folder, ])

				# add list of installed plugins
				plugins = []
				plugin_folder = settings.global_get_basefolder("plugins")
				for key, plugin in plugin_manager.plugins.items():
					if plugin.bundled or (isinstance(plugin.origin, FolderOrigin) and plugin.origin.folder == plugin_folder):
						# ignore anything bundled or from the plugins folder we already include in the backup
						continue

					plugins.append(dict(key=plugin.key,
					                    name=plugin.name,
					                    url=plugin.url))

				if len(plugins):
					zip.writestr("plugin_list.json", json.dumps(plugins))

			shutil.move(temporary_path, final_path)

			if callable(on_backup_done):
				on_backup_done(name, final_path, exclude)
		except Exception as exc:  ## noqa: F841
			# TODO py3: use the exception, not sys.exc_info()
			if callable(on_backup_error):
				exc_info = sys.exc_info()
				try:
					on_backup_error(name, exc_info)
				finally:
					del exc_info
			raise

	@classmethod
	def _restore_backup(cls, path,
	                    settings=None,
	                    plugin_manager=None,
	                    datafolder=None,
	                    on_install_plugins=None,
	                    on_report_unknown_plugins=None,
	                    on_invalid_backup=None,
	                    on_log_progress=None,
	                    on_log_error=None,
	                    on_restore_start=None,
	                    on_restore_done=None,
	                    on_restore_failed=None):
		if not is_os_compatible(["!windows"]):
			if callable(on_log_error):
				on_log_error("Restore is not supported on this operating system")
			if callable(on_restore_failed):
				on_restore_failed(path)
			return False

		restart_command = settings.global_get(["server", "commands", "serverRestartCommand"])

		basedir = settings._basedir
		cls._clean_dir_backup(basedir,
		                       on_log_progress=on_log_progress)

		plugin_repo = dict()
		repo_url = settings.global_get(["plugins", "pluginmanager", "repository"])
		if repo_url:
			plugin_repo = cls._get_plugin_repository_data(repo_url)

		if callable(on_restore_start):
			on_restore_start(path)

		try:

			with zipfile.ZipFile(path, "r") as zip:
				# read metadata
				try:
					metadata_zipinfo = zip.getinfo("metadata.json")
				except KeyError:
					if callable(on_invalid_backup):
						on_invalid_backup("Not an OctoPrint backup, lacks metadata.json")
					if callable(on_restore_failed):
						on_restore_failed(path)
					return False

				metadata_bytes = zip.read(metadata_zipinfo)
				metadata = json.loads(metadata_bytes)

				backup_version = get_comparable_version(metadata["version"], base=True)
				if backup_version > get_octoprint_version(base=True):
					if callable(on_invalid_backup):
						on_invalid_backup("Backup is from a newer version of OctoPrint and cannot be applied")
					if callable(on_restore_failed):
						on_restore_failed(path)
					return False

				# unzip to temporary folder
				temp = tempfile.mkdtemp()
				try:
					if callable(on_log_progress):
						on_log_progress("Unpacking backup to {}...".format(temp))
					abstemp = os.path.abspath(temp)
					for member in zip.infolist():
						abspath = os.path.abspath(os.path.join(temp, member.filename))
						if abspath.startswith(abstemp):
							zip.extract(member, temp)

					# sanity check
					configfile = os.path.join(temp, "basedir", "config.yaml")
					if not os.path.exists(configfile):
						if callable(on_invalid_backup):
							on_invalid_backup("Backup lacks config.yaml")
						if callable(on_restore_failed):
							on_restore_failed(path)
						return False

					import yaml

					with io.open(configfile, "rt", encoding="utf-8") as f:
						configdata = yaml.safe_load(f)

					if configdata.get("accessControl", dict()).get("enabled", True):
						userfile = os.path.join(temp, "basedir", "users.yaml")
						if not os.path.exists(userfile):
							if callable(on_invalid_backup):
								on_invalid_backup("Backup lacks users.yaml")
							if callable(on_restore_failed):
								on_restore_failed(path)
							return False

					if callable(on_log_progress):
						on_log_progress("Unpacked")

					# install available plugins
					plugins = []
					plugin_list_file = os.path.join(temp, "plugin_list.json")
					if os.path.exists(plugin_list_file):
						with io.open(os.path.join(temp, "plugin_list.json"), 'rb') as f:
							plugins = json.load(f)

					known_plugins = []
					unknown_plugins = []
					if plugins:
						if plugin_repo:
							for plugin in plugins:
								if plugin["key"] in plugin_manager.plugins:
									# already installed
									continue

								if plugin["key"] in plugin_repo:
									# not installed, can be installed from repository url
									known_plugins.append(plugin_repo[plugin["key"]])
								else:
									# not installed, not installable
									unknown_plugins.append(plugin)

						else:
							# no repo, all plugins are not installable
							unknown_plugins = plugins

						if callable(on_log_progress):
							if known_plugins:
								on_log_progress("Known and installable plugins: {}".format(", ".join(map(lambda x: x["id"], known_plugins))))
							if unknown_plugins:
								on_log_progress("Unknown plugins: {}".format(", ".join(map(lambda x: x["key"], unknown_plugins))))

						if callable(on_install_plugins):
							on_install_plugins(known_plugins)

						if callable(on_report_unknown_plugins):
							on_report_unknown_plugins(unknown_plugins)

					# move config data
					basedir_backup = basedir + ".bck"
					basedir_extracted = os.path.join(temp, "basedir")

					if callable(on_log_progress):
						on_log_progress("Renaming {} to {}...".format(basedir, basedir_backup))
					shutil.move(basedir, basedir_backup)

					try:
						if callable(on_log_progress):
							on_log_progress("Moving {} to {}...".format(basedir_extracted, basedir))
						shutil.move(basedir_extracted, basedir)
					except Exception:
						if callable(on_log_error):
							on_log_error("Error while restoring config data", exc_info=sys.exc_info())
							on_log_error("Rolling back old config data")

						shutil.move(basedir_backup, basedir)

						if callable(on_restore_failed):
							on_restore_failed(path)
						return False

					if unknown_plugins:
						if callable(on_log_progress):
							on_log_progress("Writing info file about unknown plugins")

						if not os.path.isdir(datafolder):
							os.makedirs(datafolder)

						unknown_plugins_path = os.path.join(datafolder, UNKNOWN_PLUGINS_FILE)
						try:
							with io.open(unknown_plugins_path, mode='wb') as f:
								f.write(to_bytes(json.dumps(unknown_plugins)))
						except Exception:
							if callable(on_log_error):
								on_log_error("Could not persist list of unknown plugins to {}".format(unknown_plugins_path),
								             exc_info = sys.exc_info())

				finally:
					if callable(on_log_progress):
						on_log_progress("Removing temporary unpacked folder")
					shutil.rmtree(temp)

		except Exception:
			exc_info = sys.exc_info()
			try:
				if callable(on_log_error):
					on_log_error("Error while running restore", exc_info=exc_info)
				if callable(on_restore_failed):
					on_restore_failed(path)
			finally:
				del exc_info
			return False

		finally:
			# remove zip
			if callable(on_log_progress):
				on_log_progress("Removing temporary zip")
			os.remove(path)

		# restart server
		if restart_command:
			import sarge

			if callable(on_log_progress):
				on_log_progress("Restarting...")
			if callable(on_restore_done):
				on_restore_done(path)

			try:
				sarge.run(restart_command, async_=True)
			except Exception:
				if callable(on_log_error):
					on_log_error("Error while restarting via command {}".format(restart_command),
					             exc_info=sys.exc_info())
					on_log_error("Please restart OctoPrint manually")
				return False

		else:
			if callable(on_restore_done):
				on_restore_done(path)

		return True


	def _send_client_message(self, message, payload=None):
		if payload is None:
			payload = dict()
		payload["type"] = message
		self._plugin_manager.send_plugin_message(self._identifier, payload)



class InsufficientSpace(Exception):
	pass

__plugin_name__ = "Backup & Restore"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Backup & restore your OctoPrint settings and data"
__plugin_disabling_discouraged__ = gettext("Without this plugin you will no longer be able to backup "
                                           "& restore your OctoPrint settings and data.")
__plugin_license__ = "AGPLv3"
__plugin_implementation__ = BackupPlugin()
__plugin_hooks__ = {
	"octoprint.server.http.routes": __plugin_implementation__.route_hook,
	"octoprint.server.http.bodysize": __plugin_implementation__.bodysize_hook,
	"octoprint.cli.commands": __plugin_implementation__.cli_commands_hook,
	"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
# NO unicode_literals because Py2 setuptool can't cope with them

__author__ = u"Gina Häußge <osd@foosel.net>"
__license__ = u'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = u"Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import os
import shutil
import glob

from setuptools import Command
from distutils.command.clean import clean as _clean

def package_data_dirs(source, sub_folders):
	dirs = []

	for d in sub_folders:
		folder = os.path.join(source, d)
		if not os.path.exists(folder):
			continue

		for dirname, _, files in os.walk(folder):
			dirname = os.path.relpath(dirname, source)
			for f in files:
				dirs.append(os.path.join(dirname, f))

	return dirs


def recursively_handle_files(directory, file_matcher, folder_matcher=None, folder_handler=None, file_handler=None):
	applied_handler = False

	for filename in os.listdir(directory):
		path = os.path.join(directory, filename)

		if file_handler is not None and file_matcher(filename):
			file_handler(path)
			applied_handler = True

		elif os.path.isdir(path) and (folder_matcher is None or folder_matcher(directory, filename, path)):
			sub_applied_handler = recursively_handle_files(path, file_matcher, folder_handler=folder_handler, file_handler=file_handler)
			if sub_applied_handler:
				applied_handler = True

			if folder_handler is not None:
				folder_handler(path, sub_applied_handler)

	return applied_handler


def has_requirement(requirement, requirements):
	if requirement is None or requirements is None:
		return False

	#from past.builtins import basestring

	#assert isinstance(requirement, basestring)
	#assert isinstance(requirements, (list, tuple))
	#assert all(list(map(lambda x: x is not None and isinstance(x, basestring), requirements)))

	requirement = requirement.lower()
	requirements = [r.lower() for r in requirements]
	compat = [requirement.lower() + c for c in ("<", "<=", "!=", "==", ">=", ">", "~=", "===")]

	return requirement in requirements or \
	       any(any(r.startswith(c) for c in compat) for r in requirements)


class CleanCommand(_clean):
	user_options = _clean.user_options + [("orig", None, "behave like original clean command"),
	                                      ("noeggs", None, "don't clean up eggs"),
	                                      ("nopyc", None, "don't clean up pyc files")]
	boolean_options = _clean.boolean_options + ["orig", "noeggs", "nopyc"]

	source_folder = "src"
	eggs =None

	@classmethod
	def for_options(cls, source_folder="src", eggs=None):
		if eggs is None:
			eggs = []
		return type(cls)(cls.__name__, (cls,), dict(
			source_folder=source_folder,
			eggs=eggs
		))

	def initialize_options(self):
		_clean.initialize_options(self)

		self.orig = None
		self.noeggs = None
		self.nopyc = None

	def finalize_options(self):
		_clean.finalize_options(self)

		if not self.orig:
			self.all = True

	def run(self):
		_clean.run(self)
		if self.orig:
			return

		# eggs
		if not self.noeggs:
			for egg in self.eggs:
				globbed_eggs = glob.glob(egg)
				for globbed_egg in globbed_eggs:
					print("deleting '%s' egg" % globbed_egg)
					if not self.dry_run:
						shutil.rmtree(globbed_egg)

		# pyc files
		if not self.nopyc:
			def delete_folder_if_empty(path, applied_handler):
				if not applied_handler:
					return
				if len(os.listdir(path)) == 0:
					if not self.dry_run:
						shutil.rmtree(path)
					print("removed %s since it was empty" % path[len(self.source_folder):])

			def delete_file(path):
				print("removing '%s'" % path[len(self.source_folder):])
				if not self.dry_run:
					os.remove(path)

			import fnmatch
			print("recursively removing *.pyc from '%s'" % self.source_folder)
			recursively_handle_files(
				os.path.abspath(self.source_folder),
				lambda name: fnmatch.fnmatch(name.lower(), "*.pyc"),
				folder_matcher=lambda dir, name, path: name != ".git",
				folder_handler=delete_folder_if_empty,
				file_handler=delete_file
			)


class NewTranslation(Command):
	description = "create a new translation"
	user_options = [
		('locale=', 'l', 'locale for the new translation'),
		]
	boolean_options = []

	pot_file = None
	output_dir = None

	@classmethod
	def for_options(cls, pot_file=None, output_dir=None):
		if pot_file is None:
			raise ValueError("pot_file must not be None")
		if output_dir is None:
			raise ValueError("output_dir must not be None")

		return type(cls)(cls.__name__, (cls,), dict(
			pot_file=pot_file,
			output_dir=output_dir
		))

	def __init__(self, dist, **kw):
		from babel.messages import frontend as babel
		self.babel_init_messages = babel.init_catalog(dist)
		Command.__init__(self, dist, **kw)

	def initialize_options(self):
		self.locale = None
		self.babel_init_messages.initialize_options()

	def finalize_options(self):
		self.babel_init_messages.locale = self.locale
		self.babel_init_messages.input_file = self.__class__.pot_file
		self.babel_init_messages.output_dir = self.__class__.output_dir
		self.babel_init_messages.finalize_options()

	def run(self):
		self.babel_init_messages.run()


class ExtractTranslation(Command):
	description = "extract translations"
	user_options = []
	boolean_options = []

	mail_address = "i18n@octoprint.org"
	copyright_holder = "The OctoPrint Project"
	mapping_file = None
	pot_file = None
	input_dirs = None

	@classmethod
	def for_options(cls, mail_address="i18n@octoprint.org", copyright_holder="The OctoPrint Project", mapping_file=None, pot_file=None, input_dirs=None):
		if mapping_file is None:
			raise ValueError("mapping_file must not be None")
		if pot_file is None:
			raise ValueError("pot_file must not be None")
		if input_dirs is None:
			raise ValueError("input_dirs must not be None")

		return type(cls)(cls.__name__, (cls,), dict(
			mapping_file=mapping_file,
			pot_file=pot_file,
			input_dirs=input_dirs,
			mail_address=mail_address,
			copyright_holder=copyright_holder
		))

	def __init__(self, dist, **kw):
		from babel.messages import frontend as babel
		self.babel_extract_messages = babel.extract_messages(dist)
		Command.__init__(self, dist, **kw)

	def initialize_options(self):
		self.babel_extract_messages.initialize_options()

	def finalize_options(self):
		self.babel_extract_messages.mapping_file = self.__class__.mapping_file
		self.babel_extract_messages.output_file = self.__class__.pot_file
		self.babel_extract_messages.input_dirs = self.__class__.input_dirs
		self.babel_extract_messages.msgid_bugs_address = self.__class__.mail_address
		self.babel_extract_messages.copyright_holder = self.__class__.copyright_holder
		self.babel_extract_messages.finalize_options()

	def run(self):
		self.babel_extract_messages.run()


class RefreshTranslation(Command):
	description = "refresh translations"
	user_options = [
		('locale=', 'l', 'locale for the translation to refresh'),
		]
	boolean_options = []

	mail_address = "i18n@octoprint.org"
	copyright_holder = "The OctoPrint Project"
	mapping_file = None
	pot_file = None
	input_dirs = None
	output_dir = None

	@classmethod
	def for_options(cls, mail_address="i18n@octoprint.org", copyright_holder="The OctoPrint Project", mapping_file=None, pot_file=None, input_dirs=None, output_dir=None):
		if mapping_file is None:
			raise ValueError("mapping_file must not be None")
		if pot_file is None:
			raise ValueError("pot_file must not be None")
		if input_dirs is None:
			raise ValueError("input_dirs must not be None")
		if output_dir is None:
			raise ValueError("output_dir must not be None")

		return type(cls)(cls.__name__, (cls,), dict(
			mapping_file=mapping_file,
			pot_file=pot_file,
			input_dirs=input_dirs,
			mail_address=mail_address,
			copyright_holder=copyright_holder,
			output_dir=output_dir
		))

	def __init__(self, dist, **kw):
		from babel.messages import frontend as babel
		self.babel_extract_messages = babel.extract_messages(dist)
		self.babel_update_messages = babel.update_catalog(dist)
		Command.__init__(self, dist, **kw)

	def initialize_options(self):
		self.locale = None
		self.babel_extract_messages.initialize_options()
		self.babel_update_messages.initialize_options()

	def finalize_options(self):
		self.babel_extract_messages.mapping_file = self.__class__.mapping_file
		self.babel_extract_messages.output_file = self.__class__.pot_file
		self.babel_extract_messages.input_dirs = self.__class__.input_dirs
		self.babel_extract_messages.msgid_bugs_address = self.__class__.mail_address
		self.babel_extract_messages.copyright_holder = self.__class__.copyright_holder
		self.babel_extract_messages.finalize_options()

		self.babel_update_messages.input_file = self.__class__.pot_file
		self.babel_update_messages.output_dir = self.__class__.output_dir
		self.babel_update_messages.locale = self.locale
		self.babel_update_messages.finalize_options()

	def run(self):
		self.babel_extract_messages.run()
		self.babel_update_messages.run()


class CompileTranslation(Command):
	description = "compile translations"
	user_options = []
	boolean_options = []

	output_dir = None

	@classmethod
	def for_options(cls, output_dir=None):
		if output_dir is None:
			raise ValueError("output_dir must not be None")

		return type(cls)(cls.__name__, (cls,), dict(
			output_dir=output_dir
		))

	def __init__(self, dist, **kw):
		from babel.messages import frontend as babel
		self.babel_compile_messages = babel.compile_catalog(dist)
		Command.__init__(self, dist, **kw)

	def initialize_options(self):
		self.babel_compile_messages.initialize_options()

	def finalize_options(self):
		self.babel_compile_messages.directory = self.__class__.output_dir
		self.babel_compile_messages.finalize_options()

	def run(self):
		self.babel_compile_messages.run()


class BundleTranslation(Command):
	description = "bundles translations"
	user_options = [
		('locale=', 'l', 'locale for the translation to bundle')
	]
	boolean_options = []

	source_dir = None
	target_dir = None

	@classmethod
	def for_options(cls, source_dir=None, target_dir=None):
		if source_dir is None:
			raise ValueError("source_dir must not be None")
		if target_dir is None:
			raise ValueError("target_dir must not be None")

		return type(cls)(cls.__name__, (cls,), dict(
			source_dir=source_dir,
			target_dir=target_dir
		))

	def initialize_options(self):
		self.locale = None

	def finalize_options(self):
		pass

	def run(self):
		locale = self.locale
		source_path = os.path.join(self.__class__.source_dir, locale)
		target_path = os.path.join(self.__class__.target_dir, locale)

		if not os.path.exists(source_path):
			raise RuntimeError("source path " + source_path + " does not exist")

		if os.path.exists(target_path):
			if not os.path.isdir(target_path):
				raise RuntimeError("target path " + target_path + " exists and is not a directory")
			shutil.rmtree(target_path)

		print("Copying translations for locale {locale} from {source_path} to {target_path}...".format(**locals()))
		shutil.copytree(source_path, target_path)


class PackTranslation(Command):
	description = "creates language packs for translations"
	user_options = [
		('locale=', 'l', 'locale for the translation to pack'),
		('author=', 'a', 'author of the translation'),
		('target=', 't', 'target folder for the pack')
	]
	boolean_options = []

	source_dir = None
	pack_name_prefix = None
	pack_path_prefix = None

	@classmethod
	def for_options(cls, source_dir=None, pack_name_prefix=None, pack_path_prefix=None):
		if source_dir is None:
			raise ValueError("source_dir must not be None")
		if pack_name_prefix is None:
			raise ValueError("pack_name_prefix must not be None")
		if pack_path_prefix is None:
			raise ValueError("pack_path_prefix must not be None")

		return type(cls)(cls.__name__, (cls,), dict(
			source_dir=source_dir,
			pack_name_prefix=pack_name_prefix,
			pack_path_prefix=pack_path_prefix
		))

	def initialize_options(self):
		self.locale = None
		self.author = None
		self.target = None

	def finalize_options(self):
		if self.locale is None:
			raise ValueError("locale must be provided")

	def run(self):
		locale = self.locale
		locale_dir = os.path.join(self.__class__.source_dir, locale)

		if not os.path.isdir(locale_dir):
			raise RuntimeError("translation does not exist, please create it first")

		import datetime

		now = datetime.datetime.utcnow().replace(microsecond=0)

		if self.target is None:
			self.target = self.__class__.source_dir

		zip_path = os.path.join(self.target, "{prefix}{locale}_{date}.zip".format(prefix=self.__class__.pack_name_prefix, locale=locale, date=now.strftime("%Y%m%d%H%M%S")))
		print("Packing translation to {zip_path}".format(**locals()))

		def add_recursively(zip, path, prefix):
			if not os.path.isdir(path):
				return

			for entry in os.listdir(path):
				entry_path = os.path.join(path, entry)
				new_prefix = prefix + "/" + entry
				if os.path.isdir(entry_path):
					add_recursively(zip, entry_path, new_prefix)
				elif os.path.isfile(entry_path):
					print("Adding {entry_path} as {new_prefix}".format(**locals()))
					zip.write(entry_path, new_prefix)

		meta_str = "last_update: {date}\n".format(date=now.isoformat())
		if self.author:
			meta_str += "author: {author}\n".format(author=self.author)

		zip_locale_root = self.__class__.pack_path_prefix + locale

		import zipfile
		with zipfile.ZipFile(zip_path, "w") as zip:
			add_recursively(zip, locale_dir, zip_locale_root)

			print("Adding meta.yaml as {zip_locale_root}/meta.yaml".format(**locals()))
			zip.writestr(zip_locale_root + "/meta.yaml", meta_str)


def get_babel_commandclasses(pot_file=None,
                             mapping_file="babel.cfg",
                             input_dirs=".",
                             output_dir=None,
                             pack_name_prefix=None,
                             pack_path_prefix=None,
                             bundled_dir=None,
                             mail_address="i18n@octoprint.org",
                             copyright_holder="The OctoPrint Project"):
	result = dict(
		babel_new=NewTranslation.for_options(pot_file=pot_file, output_dir=output_dir),
		babel_extract=ExtractTranslation.for_options(mapping_file=mapping_file, pot_file=pot_file, input_dirs=input_dirs, mail_address=mail_address, copyright_holder=copyright_holder),
		babel_refresh=RefreshTranslation.for_options(mapping_file=mapping_file, pot_file=pot_file, input_dirs=input_dirs, output_dir=output_dir, mail_address=mail_address, copyright_holder=copyright_holder),
		babel_compile=CompileTranslation.for_options(output_dir=output_dir),
		babel_pack=PackTranslation.for_options(source_dir=output_dir, pack_name_prefix=pack_name_prefix, pack_path_prefix=pack_path_prefix)
	)

	if bundled_dir is not None:
		result["babel_bundle"] = BundleTranslation.for_options(source_dir=output_dir, target_dir=bundled_dir)

	return result


def create_plugin_setup_parameters(identifier="todo", name="TODO", version="0.1", description="TODO", author="TODO",
                                   mail="todo@example.com", url="TODO", license="AGPLv3", source_folder=".", additional_data=None,
                                   additional_packages=None, ignored_packages=None, requires=None, extra_requires=None,
                                   cmdclass=None, eggs=None, package=None, dependency_links=None):
	import pkg_resources

	if package is None:
		package = "octoprint_{identifier}".format(**locals())

	if additional_data is None:
		additional_data = list()

	if additional_packages is None:
		additional_packages = list()

	if ignored_packages is None:
		ignored_packages = list()

	if dependency_links is None:
		dependency_links = list()

	if requires is None:
		requires = ["OctoPrint"]
	if not isinstance(requires, list):
		raise ValueError("requires must be a list")
	if not has_requirement("OctoPrint", requires):
		requires = ["OctoPrint"] + list(requires)

	if extra_requires is None:
		extra_requires = dict()
	if not isinstance(extra_requires, dict):
		raise ValueError("extra_requires must be a dict")

	if cmdclass is None:
		cmdclass = dict()
	if not isinstance(cmdclass, dict):
		raise ValueError("cmdclass must be a dict")

	if eggs is None:
		eggs = []
	if not isinstance(eggs, list):
		raise ValueError("eggs must be a list")

	egg = "{name}*.egg-info".format(name=pkg_resources.to_filename(pkg_resources.safe_name(name)))
	if egg not in eggs:
		eggs = [egg] + eggs

	cmdclass.update(dict(
		clean=CleanCommand.for_options(source_folder=os.path.join(source_folder, package), eggs=eggs)
	))

	translation_dir = os.path.join(source_folder, "translations")
	pot_file = os.path.join(translation_dir, "messages.pot")
	bundled_dir = os.path.join(source_folder, package, "translations")
	cmdclass.update(get_babel_commandclasses(pot_file=pot_file, output_dir=translation_dir, bundled_dir=bundled_dir, pack_name_prefix="{name}-i18n-".format(**locals()), pack_path_prefix="_plugins/{identifier}/".format(**locals())))

	from setuptools import find_packages
	packages = set([package]
	               + list(filter(lambda x: x.startswith("{package}.".format(package=package)),
	                             find_packages(where=source_folder, exclude=ignored_packages)))
	               + additional_packages)
	print("Found packages: {packages!r}".format(**locals()))

	return dict(
		name=name,
		version=version,
		description=description,
		author=author,
		author_email=mail,
		url=url,
		license=license,

		# adding new commands
		cmdclass=cmdclass,

		# we only have our plugin package to install
		packages=packages,

		# we might have additional data files in sub folders that need to be installed too
		package_data={package: package_data_dirs(os.path.join(source_folder, package), ["static", "templates", "translations"] + additional_data)},
		include_package_data=True,

		# If you have any package data that needs to be accessible on the file system, such as templates or static assets
		# this plugin is not zip_safe.
		zip_safe=False,

		install_requires=requires,
		extras_require=extra_requires,
		dependency_links=dependency_links,

		# Hook the plugin into the "octoprint.plugin" entry point, mapping the plugin_identifier to the plugin_package.
		# That way OctoPrint will be able to find the plugin and load it.
		entry_points={
			"octoprint.plugin": ["{identifier} = {package}".format(**locals())]
		}
	)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import base64
import sys
PY3 = sys.version_info[0] == 3

from octoprint.settings import settings
import octoprint.timelapse
import octoprint.server

from octoprint.util import deprecated, to_unicode
from octoprint.plugin import plugin_manager

import flask as _flask
import flask_login
import octoprint.vendor.flask_principal as flask_principal
import logging

from . import flask
from . import sockjs
from . import tornado
from . import watchdog


@deprecated("API keys are no longer needed for anonymous access and thus this is now obsolete")
def enforceApiKeyRequestHandler():
	pass

apiKeyRequestHandler = deprecated("apiKeyRequestHandler has been renamed to enforceApiKeyRequestHandler")(enforceApiKeyRequestHandler)


def loginFromApiKeyRequestHandler():
	"""
	``before_request`` handler for blueprints which creates a login session for the provided api key (if available)

	App session keys are handled as anonymous keys here and ignored.
	"""
	try:
		if loginUserFromApiKey():
			_flask.g.login_via_apikey = True
	except InvalidApiKeyException:
		return _flask.make_response("Invalid API key", 403)


def loginFromAuthorizationHeaderRequestHandler():
	"""
	``before_request`` handler for creating login sessions based on the Authorization header.
	"""
	try:
		if loginUserFromAuthorizationHeader():
			_flask.g.login_via_header = True
	except InvalidApiKeyException:
		return _flask.make_response("Invalid API key", 403)


class InvalidApiKeyException(Exception): pass


def loginUserFromApiKey():
	apikey = get_api_key(_flask.request)

	if not apikey:
		return False

	user = get_user_for_apikey(apikey)
	if user is None:
		# invalid API key = no API key
		return False

	if not loginUser(user):
		return False

	return True


def loginUserFromAuthorizationHeader():
	authorization_header = get_authorization_header(_flask.request)
	user = get_user_for_authorization_header(authorization_header)
	return loginUser(user)


def loginUser(user, remember=False):
	"""
	Logs the provided ``user`` into Flask Login and Principal if not None and active

	Args:
		user: the User to login. May be None in which case the login will fail
		remember: Whether to set the ``remember`` flag on the Flask Login operation

	Returns: (bool) True if the login succeeded, False otherwise

	"""
	if user is not None and user.is_active and flask_login.login_user(user, remember=remember):
		flask_principal.identity_changed.send(_flask.current_app._get_current_object(),
		                                      identity=flask_principal.Identity(user.get_id()))
		return True
	return False


def requireLoginRequestHandler():
	if _flask.request.endpoint.endswith(".static"):
		return

	if not (octoprint.server.userManager.enabled and octoprint.server.userManager.has_been_customized()):
		return

	user = flask_login.current_user
	if user is None or user.is_anonymous or not user.is_active:
		return flask.make_response("Forbidden", 403)


def corsRequestHandler():
	"""
	``before_request`` handler for blueprints which sets CORS headers for OPTIONS requests if enabled
	"""
	if _flask.request.method == 'OPTIONS' and settings().getBoolean(["api", "allowCrossOrigin"]):
		# reply to OPTIONS request for CORS headers
		return optionsAllowOrigin(_flask.request)


def corsResponseHandler(resp):
	"""
	``after_request`` handler for blueprints for which CORS is supported.

	Sets ``Access-Control-Allow-Origin`` headers for ``Origin`` request header on response.
	"""

	# Allow crossdomain
	allowCrossOrigin = settings().getBoolean(["api", "allowCrossOrigin"])
	if _flask.request.method != 'OPTIONS' and 'Origin' in _flask.request.headers and allowCrossOrigin:
		resp.headers['Access-Control-Allow-Origin'] = _flask.request.headers['Origin']

	return resp


def noCachingResponseHandler(resp):
	"""
	``after_request`` handler for blueprints which shall set no caching headers
	on their responses.

	Sets ``Cache-Control``, ``Pragma`` and ``Expires`` headers accordingly
	to prevent all client side caching from taking place.
	"""

	return flask.add_non_caching_response_headers(resp)


def noCachingExceptGetResponseHandler(resp):
	"""
	``after_request`` handler for blueprints which shall set no caching headers
	on their responses to any requests that are not sent with method ``GET``.

	See :func:`noCachingResponseHandler`.
	"""

	if _flask.request.method == "GET":
		return flask.add_no_max_age_response_headers(resp)
	else:
		return flask.add_non_caching_response_headers(resp)


def optionsAllowOrigin(request):
	"""
	Shortcut for request handling for CORS OPTIONS requests to set CORS headers.
	"""

	resp = _flask.current_app.make_default_options_response()

	# Allow the origin which made the XHR
	resp.headers['Access-Control-Allow-Origin'] = request.headers['Origin']
	# Allow the actual method
	resp.headers['Access-Control-Allow-Methods'] = request.headers['Access-Control-Request-Method']
	# Allow for 10 seconds
	resp.headers['Access-Control-Max-Age'] = "10"

	# 'preflight' request contains the non-standard headers the real request will have (like X-Api-Key)
	customRequestHeaders = request.headers.get('Access-Control-Request-Headers', None)
	if customRequestHeaders is not None:
		# If present => allow them all
		resp.headers['Access-Control-Allow-Headers'] = customRequestHeaders

	return resp


def get_user_for_apikey(apikey):
	if apikey is not None:
		if apikey == settings().get(["api", "key"]):
			# master key was used
			return octoprint.server.userManager.api_user_factory()

		if octoprint.server.userManager.enabled:
			user = octoprint.server.userManager.find_user(apikey=apikey)
			if user is not None:
				# user key was used
				return user

		apikey_hooks = plugin_manager().get_hooks("octoprint.accesscontrol.keyvalidator")
		for name, hook in apikey_hooks.items():
			try:
				user = hook(apikey)
				if user is not None:
					return user
			except Exception:
				logging.getLogger(__name__).exception("Error running api key validator "
				                                      "for plugin {} and key {}".format(name, apikey),
				                                      extra=dict(plugin=name))
	return None


def get_user_for_remote_user_header(request):
	if not octoprint.server.userManager.enabled:
		return None

	if not settings().getBoolean(["accessControl", "trustRemoteUser"]):
		return None

	header = request.headers.get(settings().get(["accessControl", "remoteUserHeader"]))
	if header is None:
		return None

	user = octoprint.server.userManager.findUser(userid=header)

	if user is None and settings().getBoolean(["accessControl", "addRemoteUsers"]):
		octoprint.server.userManager.addUser(header, settings().generateApiKey(), active=True)
		user = octoprint.server.userManager.findUser(userid=header)

	return user


def get_user_for_authorization_header(header):
	if not settings().getBoolean(["accessControl", "trustBasicAuthentication"]):
		return None

	if header is None:
		return None

	if not header.startswith("Basic "):
		# we currently only support Basic Authentication
		return None

	header = header.replace('Basic ', '', 1)
	try:
		header = to_unicode(base64.b64decode(header))
	except TypeError:
		return None

	name, password = header.split(':', 1)
	if not octoprint.server.userManager.enabled:
		return None

	user = octoprint.server.userManager.find_user(userid=name)
	if settings().getBoolean(["accessControl", "checkBasicAuthenticationPassword"]) \
			and not octoprint.server.userManager.check_password(name, password):
		# password check enabled and password don't match
		return None

	return user


def get_api_key(request):
	# Check Flask GET/POST arguments
	if hasattr(request, "values") and "apikey" in request.values:
		return request.values["apikey"]

	# Check Tornado GET/POST arguments
	if hasattr(request, "arguments") and "apikey" in request.arguments \
		and len(request.arguments["apikey"]) > 0 and len(request.arguments["apikey"].strip()) > 0:
		return request.arguments["apikey"]

	# Check Tornado and Flask headers
	if "X-Api-Key" in request.headers:
		return request.headers.get("X-Api-Key")

	return None


def get_authorization_header(request):
	# Tornado and Flask headers
	return request.headers.get("Authorization")


def get_plugin_hash():
	from octoprint.plugin import plugin_manager

	plugin_signature = lambda impl: "{}:{}".format(impl._identifier, impl._plugin_version)
	template_plugins = list(map(plugin_signature, plugin_manager().get_implementations(octoprint.plugin.TemplatePlugin)))
	asset_plugins = list(map(plugin_signature, plugin_manager().get_implementations(octoprint.plugin.AssetPlugin)))
	ui_plugins = sorted(set(template_plugins + asset_plugins))

	import hashlib
	plugin_hash = hashlib.sha1()
	plugin_hash.update(",".join(ui_plugins).encode('utf-8'))
	return plugin_hash.hexdigest()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
    sockjs.tornado.static
    ~~~~~~~~~~~~~~~~~~~~~

    Various static handlers required for SockJS to function properly.
"""

import time
import hashlib
import random
import sys

from tornado.web import asynchronous

from octoprint.vendor.sockjs.tornado.basehandler import BaseHandler, PreflightHandler
from octoprint.vendor.sockjs.tornado.proto import json_encode
from octoprint.vendor.sockjs.tornado.util import MAXSIZE, str_to_bytes

IFRAME_TEXT = '''<!DOCTYPE html>
<html>
<head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script>
    document.domain = document.domain;
    _sockjs_onload = function(){SockJS.bootstrap_iframe();};
  </script>
  <script src="%s"></script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>'''.strip()


class IFrameHandler(BaseHandler):
    """SockJS IFrame page handler"""
    def get(self):
        data = str_to_bytes(IFRAME_TEXT % self.server.settings['sockjs_url'])

        hsh = hashlib.md5(data).hexdigest()

        value = self.request.headers.get('If-None-Match')
        if value:
            if value.find(hsh) != -1:
                # TODO: Fix me? Right now it is a hack to remove content-type
                # header
                self.clear()
                del self._headers['Content-Type']

                self.set_status(304)
                return

        self.enable_cache()

        self.set_header('Etag', hsh)
        self.write(data)


class GreetingsHandler(BaseHandler):
    """SockJS greetings page handler"""

    def initialize(self, server):
        self.server = server

    def get(self):
        self.enable_cache()

        self.set_header('Content-Type', 'text/plain; charset=UTF-8')
        self.write('Welcome to SockJS!\n')


class ChunkingTestHandler(PreflightHandler):
    """SockJS chunking test handler"""

    # Step timeouts according to sockjs documentation
    steps = [0.005, 0.025, 0.125, 0.625, 3.125]

    def initialize(self, server):
        self.server = server
        self.step = 0
        self.io_loop = server.io_loop

    @asynchronous
    def post(self):
        self.preflight()
        self.set_header('Content-Type', 'application/javascript; charset=UTF-8')

        # Send one 'h' immediately
        self.write('h\n')
        self.flush()

        # Send 2048 spaces followed by 'h'
        self.write(' ' * 2048 + 'h\n')
        self.flush()

        # Send 'h' with different timeouts
        def run_step():
            try:
                self.write('h\n')
                self.flush()

                self.step += 1
                if self.step < len(self.steps):
                    self.io_loop.add_timeout(time.time() + self.steps[self.step],
                                             run_step)
                else:
                    self.finish()
            except IOError:
                pass

        self.io_loop.add_timeout(time.time() + self.steps[0], run_step)


class InfoHandler(PreflightHandler):
    """SockJS 0.2+ /info handler"""
    def initialize(self, server):
        self.server = server
        self.access_methods = 'OPTIONS, GET'

    def get(self):
        self.preflight()
        self.disable_cache()
        self.set_header('Content-Type', 'application/json; charset=UTF-8')

        options = dict(websocket=self.server.websockets_enabled,
                       cookie_needed=self.server.cookie_needed,
                       origins=['*:*'],
                       entropy=random.randint(0, MAXSIZE))

        self.write(json_encode(options))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from . import commandline, git_commit, github_commit, github_release, bitbucket_commit, python_checker, never_current, always_current, pypi_release

def log_github_ratelimit(logger, r):
	ratelimit = r.headers["X-RateLimit-Limit"] if "X-RateLimit-Limit" in r.headers else "?"
	remaining = r.headers["X-RateLimit-Remaining"] if "X-RateLimit-Remaining" in r.headers else "?"
	reset = r.headers["X-RateLimit-Reset"] if "X-RateLimit-Reset" in r.headers else None
	try:
		import time
		reset = time.strftime("%Y-%m-%d %H:%M", time.gmtime(int(reset)))
	except Exception:
		reset = "?"

	logger.debug("Github rate limit: %s/%s, reset at %s" % (remaining, ratelimit, reset))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2016 The OctoPrint Project - Released under terms of the AGPLv3 License"


import octoprint.plugin

import calendar
import io
import os
import re
import time
import threading
import sys
PY2 = sys.version_info[0] < 3

import feedparser
import flask

from collections import OrderedDict

from octoprint.access import ADMIN_GROUP
from octoprint.access.permissions import Permissions
from octoprint.server.util.flask import no_firstrun_access, with_revalidation_checking, check_etag
from octoprint.util import utmify, count, monotonic_time
from flask_babel import gettext
from octoprint import __version__ as OCTOPRINT_VERSION


class AnnouncementPlugin(octoprint.plugin.AssetPlugin,
                         octoprint.plugin.SettingsPlugin,
                         octoprint.plugin.BlueprintPlugin,
                         octoprint.plugin.StartupPlugin,
                         octoprint.plugin.TemplatePlugin,
                         octoprint.plugin.EventHandlerPlugin):

	# noinspection PyMissingConstructor
	def __init__(self):
		self._cached_channel_configs = None
		self._cached_channel_configs_mutex = threading.RLock()

		from slugify import Slugify
		self._slugify = Slugify()
		self._slugify.safe_chars = "-_."

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="READ",
			     name="Read announcements",
			     description=gettext("Allows to read announcements"),
			     default_groups=[ADMIN_GROUP],
			     roles=["read"]),
			dict(key="MANAGE",
			     name="Manage announcement subscriptions",
			     description=gettext("Allows to manage announcement subscriptions. Includes \"Read announcements\" "
			                         "permission"),
			     default_groups=[ADMIN_GROUP],
			     roles=["manage"],
			     permissions=["PLUGIN_ANNOUNCEMENTS_READ"])
		]

	# StartupPlugin

	def on_after_startup(self):
		# decouple channel fetching from server startup
		def fetch_data():
			self._fetch_all_channels()

		thread = threading.Thread(target=fetch_data)
		thread.daemon = True
		thread.start()

	# SettingsPlugin

	def get_settings_defaults(self):
		settings = dict(channels=dict(_important=dict(name="Important Announcements",
		                                              description="Important announcements about OctoPrint.",
		                                              priority=1,
		                                              type="rss",
		                                              url="https://octoprint.org/feeds/important.xml"),
		                              _releases=dict(name="Release Announcements",
		                                             description="Announcements of new releases and release candidates of OctoPrint.",
		                                             priority=2,
		                                             type="rss",
		                                             url="https://octoprint.org/feeds/releases.xml"),
		                              _blog=dict(name="On the OctoBlog",
		                                         description="Development news, community spotlights, OctoPrint On Air episodes and more from the official OctoBlog.",
		                                         priority=2,
		                                         type="rss",
		                                         url="https://octoprint.org/feeds/octoblog.xml"),
		                              _plugins=dict(name="New Plugins in the Repository",
		                                            description="Announcements of new plugins released on the official Plugin Repository.",
		                                            priority=2,
		                                            type="rss",
		                                            url="https://plugins.octoprint.org/feed.xml"),
		                              _octopi=dict(name="OctoPi News",
		                                           description="News around OctoPi, the Raspberry Pi image including OctoPrint.",
		                                           priority=2,
		                                           type="rss",
		                                           url="https://octoprint.org/feeds/octopi.xml")),
		                enabled_channels=[],
		                forced_channels=["_important"],
		                channel_order=["_important", "_releases", "_blog", "_plugins", "_octopi"],
		                ttl=6*60,
		                display_limit=3,
		                summary_limit=300)
		settings["enabled_channels"] = list(settings["channels"].keys())
		return settings

	def get_settings_version(self):
		return 1

	def on_settings_migrate(self, target, current):
		if current is None:
			# first version had different default feeds and only _important enabled by default
			channels = self._settings.get(["channels"])
			if "_news" in channels:
				del channels["_news"]
			if "_spotlight" in channels:
				del channels["_spotlight"]
			self._settings.set(["channels"], channels)

			enabled = self._settings.get(["enabled_channels"])
			add_blog = False
			if "_news" in enabled:
				add_blog = True
				enabled.remove("_news")
			if "_spotlight" in enabled:
				add_blog = True
				enabled.remove("_spotlight")
			if add_blog and not "_blog" in enabled:
				enabled.append("_blog")
			self._settings.set(["enabled_channels"], enabled)

	# AssetPlugin

	def get_assets(self):
		return dict(js=["js/announcements.js"],
		            less=["less/announcements.less"],
		            css=["css/announcements.css"])

	# Template Plugin

	def get_template_configs(self):
		return [
			dict(type="settings", name=gettext("Announcements"), template="announcements_settings.jinja2", custom_bindings=True),
			dict(type="navbar", template="announcements_navbar.jinja2", styles=["display: none"], data_bind="visible: loginState.hasPermission(access.permissions.PLUGIN_ANNOUNCEMENTS_ANNOUNCEMENT)")
		]

	# Blueprint Plugin

	@octoprint.plugin.BlueprintPlugin.route("/channels", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_ANNOUNCEMENTS_READ.require(403)
	def get_channel_data(self):
		from octoprint.settings import valid_boolean_trues

		result = []

		force = flask.request.values.get("force", "false") in valid_boolean_trues

		enabled = self._settings.get(["enabled_channels"])
		forced = self._settings.get(["forced_channels"])

		channel_configs = self._get_channel_configs(force=force)

		def view():
			channel_data = self._fetch_all_channels(force=force)

			for key, data in channel_configs.items():
				read_until = channel_configs[key].get("read_until", None)
				entries = sorted(self._to_internal_feed(channel_data.get(key, []), read_until=read_until), key=lambda e: e["published"], reverse=True)
				unread = count(filter(lambda e: not e["read"], entries))

				if read_until is None and entries:
					last = entries[0]["published"]
					self._mark_read_until(key, last)

				result.append(dict(key=key,
				                   channel=data["name"],
				                   url=data["url"],
				                   description=data.get("description", ""),
				                   priority=data.get("priority", 2),
				                   enabled=key in enabled or key in forced,
				                   forced=key in forced,
				                   data=entries,
				                   unread=unread))

			return flask.jsonify(channels=result)

		def etag():
			import hashlib
			hash = hashlib.sha1()
			def hash_update(value):
				hash.update(value.encode("utf-8"))
			hash_update(repr(sorted(enabled)))
			hash_update(repr(sorted(forced)))
			hash_update(OCTOPRINT_VERSION)

			for channel in sorted(channel_configs.keys()):
				hash_update(repr(channel_configs[channel]))
				channel_data = self._get_channel_data_from_cache(channel, channel_configs[channel])
				hash_update(repr(channel_data))

			return hash.hexdigest()

		# noinspection PyShadowingNames
		def condition(lm, etag):
			return check_etag(etag)

		return with_revalidation_checking(etag_factory=lambda *args, **kwargs: etag(),
		                                  condition=lambda lm, etag: condition(lm, etag),
		                                  unless=lambda: force)(view)()

	@octoprint.plugin.BlueprintPlugin.route("/channels/<channel>", methods=["POST"])
	@no_firstrun_access
	@Permissions.PLUGIN_ANNOUNCEMENTS_READ.require(403)
	def channel_command(self, channel):
		from octoprint.server.util.flask import get_json_command_from_request
		from octoprint.server import NO_CONTENT

		valid_commands = dict(read=["until"],
		                      toggle=[])

		command, data, response = get_json_command_from_request(flask.request, valid_commands=valid_commands)
		if response is not None:
			return response

		if command == "read":
			until = data["until"]
			self._mark_read_until(channel, until)

		elif command == "toggle":
			if not Permissions.PLUGIN_ANNOUNCEMENTS_MANAGE.can():
				return flask.make_response("Insufficient rights", 403)
			self._toggle(channel)

		return NO_CONTENT

	def is_blueprint_protected(self):
		return False

	##~~ EventHandlerPlugin

	def on_event(self, event, payload):
		from octoprint.events import Events
		if event != Events.CONNECTIVITY_CHANGED or not payload or not payload.get("new", False):
			return
		self._fetch_all_channels_async()

	# Internal Tools

	def _mark_read_until(self, channel, until):
		"""Set read_until timestamp of a channel."""

		current_read_until = None
		channel_data = self._settings.get(["channels", channel], merged=True)
		if channel_data:
			current_read_until = channel_data.get("read_until", None)

		defaults = dict(plugins=dict(announcements=dict(channels=dict())))
		defaults["plugins"]["announcements"]["channels"][channel] = dict(read_until=current_read_until)

		with self._cached_channel_configs_mutex:
			self._settings.set(["channels", channel, "read_until"], until, defaults=defaults)
			self._settings.save()
			self._cached_channel_configs = None

	def _toggle(self, channel):
		"""Toggle enable/disabled state of a channel."""

		enabled_channels = list(self._settings.get(["enabled_channels"]))

		if channel in enabled_channels:
			enabled_channels.remove(channel)
		else:
			enabled_channels.append(channel)

		self._settings.set(["enabled_channels"], enabled_channels)
		self._settings.save()

	def _get_channel_configs(self, force=False):
		"""Retrieve all channel configs with sanitized keys."""

		with self._cached_channel_configs_mutex:
			if self._cached_channel_configs is None or force:
				configs = self._settings.get(["channels"], merged=True)
				order = self._settings.get(["channel_order"])
				all_keys = order + [key for key in sorted(configs.keys()) if not key in order]

				result = OrderedDict()
				for key in all_keys:
					config = configs.get(key)
					if config is None or "url" not in config or "name" not in config:
						# strip invalid entries
						continue
					result[self._slugify(key)] = config
				self._cached_channel_configs = result
		return self._cached_channel_configs

	def _get_channel_config(self, key, force=False):
		"""Retrieve specific channel config for channel."""

		safe_key = self._slugify(key)
		return self._get_channel_configs(force=force).get(safe_key)

	def _fetch_all_channels_async(self, force=False):
		thread = threading.Thread(target=self._fetch_all_channels, kwargs=dict(force=force))
		thread.daemon = True
		thread.start()

	def _fetch_all_channels(self, force=False):
		"""Fetch all channel feeds from cache or network."""

		channels = self._get_channel_configs(force=force)
		enabled = self._settings.get(["enabled_channels"])
		forced = self._settings.get(["forced_channels"])

		all_channels = dict()
		for key, config in channels.items():
			if not key in enabled and not key in forced:
				continue

			if not "url" in config:
				continue

			data = self._get_channel_data(key, config, force=force)
			if data is not None:
				all_channels[key] = data

		return all_channels

	def _get_channel_data(self, key, config, force=False):
		"""Fetch individual channel feed from cache/network."""

		data = None

		if not force:
			# we may use the cache, see if we have something in there
			data = self._get_channel_data_from_cache(key, config)

		if data is None:
			# cache not allowed or empty, fetch from network
			if self._connectivity_checker.online:
				data = self._get_channel_data_from_network(key, config)
			else:
				self._logger.info("Looks like we are offline, can't fetch announcements for channel {} from network".format(key))

		return data

	def _get_channel_data_from_cache(self, key, config):
		"""Fetch channel feed from cache."""

		channel_path = self._get_channel_cache_path(key)

		if os.path.exists(channel_path):
			if "ttl" in config and isinstance(config["ttl"], int):
				ttl = config["ttl"]
			else:
				ttl = self._settings.get_int(["ttl"])

			ttl *= 60
			now = time.time()
			if os.stat(channel_path).st_mtime + ttl > now:
				d = feedparser.parse(channel_path)
				self._logger.debug("Loaded channel {} from cache at {}".format(key, channel_path))
				return d

		return None

	def _get_channel_data_from_network(self, key, config):
		"""Fetch channel feed from network."""

		import requests

		url = config["url"]
		try:
			start = monotonic_time()
			r = requests.get(url, timeout=30)
			r.raise_for_status()
			self._logger.info("Loaded channel {} from {} in {:.2}s".format(key, config["url"], monotonic_time() - start))
		except Exception:
			self._logger.exception(
				"Could not fetch channel {} from {}".format(key, config["url"]))
			return None

		response = r.text
		channel_path = self._get_channel_cache_path(key)
		with io.open(channel_path, mode="wt", encoding="utf-8") as f:
			f.write(response)
		return feedparser.parse(response)

	def _to_internal_feed(self, feed, read_until=None):
		"""Convert feed to internal data structure."""

		result = []
		if "entries" in feed:
			for entry in feed["entries"]:
				try:
					internal_entry = self._to_internal_entry(entry, read_until=read_until)
					if internal_entry:
						result.append(internal_entry)
				except Exception:
					self._logger.exception("Error while converting entry from feed, skipping it")
		return result

	def _to_internal_entry(self, entry, read_until=None):
		"""Convert feed entries to internal data structure."""

		timestamp = entry.get("published_parsed", None)
		if timestamp is None:
			timestamp = entry.get("updated_parsed", None)
		if timestamp is None:
			return  None

		published = calendar.timegm(timestamp)

		read = True
		if read_until is not None:
			read = published <= read_until

		return dict(title=entry["title"],
		            title_without_tags=_strip_tags(entry["title"]),
		            summary=_lazy_images(entry["summary"]),
		            summary_without_images=_strip_images(entry["summary"]),
		            published=published,
		            link=utmify(entry["link"], source="octoprint", medium="announcements", content=OCTOPRINT_VERSION),
		            read=read)

	def _get_channel_cache_path(self, key):
		"""Retrieve cache path for channel key."""

		safe_key = self._slugify(key)
		return os.path.join(self.get_plugin_data_folder(), "{}.cache".format(safe_key))


_image_tag_re = re.compile(r'<img.*?/?>')
def _strip_images(text):
	"""
	>>> _strip_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>") # doctest: +ALLOW_UNICODE
	"<a href='test.html'>I'm a link</a> and this is an image: "
	>>> _strip_images("One <img src=\\"one.jpg\\"> and two <img src='two.jpg' > and three <img src=three.jpg> and four <img src=\\"four.png\\" alt=\\"four\\">") # doctest: +ALLOW_UNICODE
	'One  and two  and three  and four '
	>>> _strip_images("No images here") # doctest: +ALLOW_UNICODE
	'No images here'
	"""
	return _image_tag_re.sub('', text)

def _replace_images(text, callback):
	"""
	>>> callback = lambda img: "foobar"
	>>> _replace_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>", callback) # doctest: +ALLOW_UNICODE
	"<a href='test.html'>I'm a link</a> and this is an image: foobar"
	>>> _replace_images("One <img src=\\"one.jpg\\"> and two <img src='two.jpg' > and three <img src=three.jpg> and four <img src=\\"four.png\\" alt=\\"four\\">", callback) # doctest: +ALLOW_UNICODE
	'One foobar and two foobar and three foobar and four foobar'
	"""
	result = text
	for match in _image_tag_re.finditer(text):
		tag = match.group(0)
		replaced = callback(tag)
		result = result.replace(tag, replaced)
	return result

_image_src_re = re.compile(r'src=(?P<quote>[\'"]*)(?P<src>.*?)(?P=quote)(?=\s+|>)')
def _lazy_images(text, placeholder=None):
	"""
	>>> _lazy_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>") # doctest: +ALLOW_UNICODE
	'<a href=\\'test.html\\'>I\\'m a link</a> and this is an image: <img src="data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7" data-src=\\'foo.jpg\\' alt=\\'foo\\'>'
	>>> _lazy_images("<a href='test.html'>I'm a link</a> and this is an image: <img src='foo.jpg' alt='foo'>", placeholder="ph.png") # doctest: +ALLOW_UNICODE
	'<a href=\\'test.html\\'>I\\'m a link</a> and this is an image: <img src="ph.png" data-src=\\'foo.jpg\\' alt=\\'foo\\'>'
	>>> _lazy_images("One <img src=\\"one.jpg\\"> and two <img src='two.jpg' > and three <img src=three.jpg> and four <img src=\\"four.png\\" alt=\\"four\\">", placeholder="ph.png") # doctest: +ALLOW_UNICODE
	'One <img src="ph.png" data-src="one.jpg"> and two <img src="ph.png" data-src=\\'two.jpg\\' > and three <img src="ph.png" data-src=three.jpg> and four <img src="ph.png" data-src="four.png" alt="four">'
	>>> _lazy_images("No images here") # doctest: +ALLOW_UNICODE
	'No images here'
	"""
	if placeholder is None:
		# 1px transparent gif
		placeholder = "data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7"

	def callback(img_tag):
		match = _image_src_re.search(img_tag)
		if match is not None:
			src = match.group("src")
			quote = match.group("quote")
			quoted_src = quote + src + quote
			img_tag = img_tag.replace(match.group(0), 'src="{}" data-src={}'.format(placeholder, quoted_src))
		return img_tag

	return _replace_images(text, callback)

def _strip_tags(text):
	"""
	>>> _strip_tags("<a href='test.html'>Hello world</a>&lt;img src='foo.jpg'&gt;") # doctest: +ALLOW_UNICODE
	"Hello world&lt;img src='foo.jpg'&gt;"
	>>> _strip_tags("&#62; &#x3E; Foo") # doctest: +ALLOW_UNICODE
	'&#62; &#x3E; Foo'
	"""
	try:
		# noinspection PyCompatibility
		from html.parser import HTMLParser
	except ImportError:
		# noinspection PyCompatibility
		from HTMLParser import HTMLParser

	class TagStripper(HTMLParser):

		def __init__(self, **kw):
			HTMLParser.__init__(self, **kw)
			self._fed = []

		def handle_data(self, data):
			self._fed.append(data)

		def handle_entityref(self, ref):
			self._fed.append("&{};".format(ref))

		def handle_charref(self, ref):
			self._fed.append("&#{};".format(ref))

		def get_data(self):
			return "".join(self._fed)

	tag_stripper = TagStripper() if PY2 else TagStripper(convert_charrefs=False)
	tag_stripper.feed(text)
	return tag_stripper.get_data()


__plugin_name__ = "Announcement Plugin"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Announcements all around OctoPrint"
__plugin_disabling_discouraged__ = gettext("Without this plugin you might miss important announcements "
                                           "regarding security or other critical issues concerning OctoPrint.")
__plugin_license__ = "AGPLv3"
__plugin_implementation__ = AnnouncementPlugin()

__plugin_hooks__ = {
	'octoprint.access.permissions': __plugin_implementation__.get_additional_permissions
}
